########################################################################
#                       AutoLigand                         
#                Author:  Rodney M. Harris                 
#                Copyright: Rodney M. Harris TSRI 2007, 2008, 2009    
#   please site:
#   Harris R, Olson A, Goodsell D.  Automated prediction
#   of ligand-binding sites in proteins.  Proteins 2008; 70:1505-1517
#########################################################################
# $Header: /opt/cvs/python/packages/share1.5/AutoDockTools/AutoLigand.py,v 1.10 2009/10/01 17:12:27 sargis Exp $
#
# $Id: AutoLigand.py,v 1.10 2009/10/01 17:12:27 sargis Exp $
import sys, string
import getopt
try:
	from Numeric import zeros
except:
	from numpy.oldnumeric import zeros
from math import sqrt
import copy
def usage():
	"Print helpful usage statement to stdout."
	print "Usage:  python AutoLigand.py -r FileBaseName -p #_of_pts"
	print
	print "    Description of command..."
	print "       -r FileBaseName = just the name part from receptor map files (i.e., FileBaseName.C.map)"
	print "       -p #_of_pts = number of fill points to use (int)"
	print "              Note: can be omitted if -a option used."
	print "    Optional parameters:"
	print "       [-a #] = number of heavy atom for ligand (#_of_pts will be set to 6x atoms)"
	print "       [-x # -y # -z #] = optional x,y,z co-ords for starting fill (float)"
	print "              when starting point is input, only one fill will be run"
	print "       [-i # -j # -k #] = optional x,y,z co-ords for second point (float)"
	print "              when second point is input, the fill will connect both points"
	print "              NOTE: the connection path has not been optimized - use with discretion"
	print "       [-f #] = number of fills to generate - default is 10"
	print "       [-e] = use the extra atom types NA, N, SA, and A"
	print "              NOTE: these results can be problematic - use with discretion"
	print "       [-m] = make a movie of output fill progress"
# process command arguments
try:
	opt_list, args = getopt.getopt(sys.argv[1:], 'r:p:a:x:y:z:i:j:k:f:em')
except getopt.GetoptError, msg:
	print 'AutoLigand.py: %s' %msg
	usage()
	sys.exit(2)
# initialize parameters
FileBaseName = None
volume = 0
x_flag = False
y_flag = False
z_flag = False
i_flag = False
j_flag = False
k_flag = False
Number_fills = 10
more_atom_types = False
movie = False
# 'r:p:a:x:y:z:i:j:k:f:xm'
for o, a in opt_list:
	if o in ('-r', '--r'):
		FileBaseName = a
	if o in ('-p', '--p'):
		volume = int(a)
	if o in ('-a', '--a'):
		no_heavy_atoms = int(a)
		volume = no_heavy_atoms * 6
	if o in ('-x', '--x'):
		x_input = float(a)
		x_flag = True
	if o in ('-y', '--y'):
		y_input = float(a)
		y_flag = True
	if o in ('-z', '--z'):
		z_input = float(a)
		z_flag = True
	if o in ('-i', '--i'):
		x2_input = float(a)
		i_flag = True
	if o in ('-j', '--j'):
		y2_input = float(a)
		j_flag = True
	if o in ('-k', '--k'):
		z2_input = float(a)
		k_flag = True
	if o in ('-f', '--f'):
		Number_fills = int(a)
	if o in ('-e', '--e'):
		more_atom_types = True
	if o in ('-m', '--m'):
		movie = True
if not FileBaseName and not volume:
	usage()
	sys.exit(2)
if not x_flag:
	start_point_flag = 0
if x_flag and y_flag and z_flag and not i_flag:
	start_point_flag = 1
if x_flag and y_flag and z_flag and i_flag and j_flag and k_flag:
	start_point_flag = 2
# GUI
if movie:
	import cPickle
	output = open(FileBaseName+'_flood.pkl', 'wb')
prev_flood = []

def save_flood(flood):
	global prev_flood
	add_set = []
	for item in flood:
		if item in prev_flood:
			prev_flood.remove(item)
		else:
			add_set.append(item)
	if prev_flood or add_set:
		cPickle.dump([prev_flood, add_set], output, -1)
	prev_flood = copy.copy(flood)

every = 3 # note, this is hard-coded - if it needs changing, the StartPointModList must be changed below
# in this version of AutoLigand, only C, OA, and HD are used - the user can also add A, N, NA, SA
Aflag = 0
Cflag = 1
Hflag = 1
Nflag = 0
NAflag = 0
Oflag = 1
Sflag = 0
CFileName = FileBaseName + '.C.map'
HFileName = FileBaseName + '.HD.map'
OFileName = FileBaseName + '.OA.map'
dFileName = FileBaseName + '.d.map'
eFileName = FileBaseName + '.e.map'
try:
	Cmap = open(CFileName,'r')
except IOError:
	Cflag = 0
try:
	Hmap = open(HFileName,'r')
except IOError:
	Hflag = 0
try:
	Omap = open(OFileName,'r')
except IOError:
	Oflag = 0
emap = open(eFileName,'r')
dmap = open(dFileName,'r')
if more_atom_types:
	Aflag = 1
	Nflag = 1
	NAflag = 1
	Sflag = 1
	AFileName = FileBaseName + '.A.map'
	NFileName = FileBaseName + '.N.map'
	NAFileName = FileBaseName + '.NA.map'
	SFileName = FileBaseName + '.SA.map'
	try:
		Amap = open(AFileName,'r')
	except IOError:
		Aflag = 0
	try:
		Nmap = open(NFileName,'r')
	except IOError:
		Nflag = 0
	try:
		NAmap = open(NAFileName,'r')
	except IOError:
		NAflag = 0
	try:
		Smap = open(SFileName,'r')
	except IOError:
		Sflag = 0
for i in range(6):
	linee = emap.readline()
	lined = dmap.readline()
	if Aflag == 1: lineA = Amap.readline()
	if Cflag == 1: lineC = Cmap.readline()
	if Hflag == 1: lineH = Hmap.readline()
	if Nflag == 1: lineN = Nmap.readline()
	if NAflag == 1: lineNA = NAmap.readline()
	if Oflag == 1: lineO = Omap.readline()
	if Sflag == 1: lineS = Smap.readline()
	if i == 3:
		word = string.split(linee)
		spacing = string.atof(word[1])
	if i == 4:
		word = string.split(linee)
		nelementsx = string.atoi(word[1])
		nelementsy = string.atoi(word[2])
		nelementsz = string.atoi(word[3])
	if i == 5:
		word = string.split(linee)
		centerx = string.atof(word[1])
		centery = string.atof(word[2])
		centerz = string.atof(word[3])
Alist = []
Clist = []
Hlist = []
Nlist = []
NAlist = []
Olist = []
Slist = []
MINlist = []
MINatom = []
# charge parameters
qA = 0.01
qC = 0.0
qH = 0.163
qN = -0.24
qNA = -0.25
qO = -0.395
qS = -0.180
nx = nelementsx + 1
ny = nelementsy + 1
nz = nelementsz + 1
total_points = nx * ny * nz
i = 0
print 'total grid points = ', total_points

# GUI
if movie:
	print "output progress file = data.pkl"
	data = [int(nx/2), int(ny/2), int(nz/2),
		    centerx, centery, centerz,
		    spacing]
	cPickle.dump(data, output, -1)

while i < total_points:
	linee = emap.readline()
	worde = string.split(linee)
	eMapVal = string.atof(worde[0])
	lined = dmap.readline()
	wordd = string.split(lined)
	dMapVal = string.atof(wordd[0])
	if Aflag == 1:
		lineA = Amap.readline()
		word = string.split(lineA)
		rawdatum = string.atof(word[0])
		# datum = string.atof(word[0])
		# datum = rawdatum + (qA * eMapVal) + (abs(qA) * dMapVal)
		datum = rawdatum + (abs(qA) * dMapVal)
		Alist.append(datum)
	else: Alist.append(1000000.)
	if Cflag == 1:
		lineC = Cmap.readline()
		word = string.split(lineC)
		rawdatum = string.atof(word[0])
		# datum = string.atof(word[0])
		# datum = rawdatum + (qC * eMapVal) + (abs(qC) * dMapVal)
		datum = rawdatum + (abs(qC) * dMapVal)
		Clist.append(datum)
	else: Clist.append(900000.)
	# the Clist is set lower than the others as a default
	if Hflag == 1:
		lineH = Hmap.readline()
		word = string.split(lineH)
		rawdatum = string.atof(word[0])
		# datum = string.atof(word[0])
		# datum = rawdatum + (qH * eMapVal) + (abs(qH) * dMapVal)
		datum = rawdatum + (abs(qH) * dMapVal)
		Hlist.append(datum)
	else: Hlist.append(1000000.)
	if Nflag == 1:
		lineN = Nmap.readline()
		word = string.split(lineN)
		rawdatum = string.atof(word[0])
		# datum = string.atof(word[0])
		# datum = rawdatum + (qN * eMapVal) + (abs(qN) * dMapVal)
		datum = rawdatum + (abs(qN) * dMapVal)
		Nlist.append(datum)
	else: Nlist.append(1000000.)
	if NAflag == 1:
		lineNA = NAmap.readline()
		word = string.split(lineNA)
		rawdatum = string.atof(word[0])
		# datum = string.atof(word[0])
		# datum = rawdatum + (qNA * eMapVal) + (abs(qNA) * dMapVal)
		datum = rawdatum + (abs(qNA) * dMapVal)
		NAlist.append(datum)
	else: NAlist.append(1000000.)
	if Oflag == 1:
		lineO = Omap.readline()
		word = string.split(lineO)
		rawdatum = string.atof(word[0])
		# datum = string.atof(word[0])
		# datum = rawdatum + (qO * eMapVal) + (abs(qO) * dMapVal)
		datum = rawdatum + (abs(qO) * dMapVal)
		Olist.append(datum)
	else: Olist.append(1000000.)
	if Sflag == 1:
		lineS = Smap.readline()
		word = string.split(lineS)
		rawdatum = string.atof(word[0])
		# datum = string.atof(word[0])
		# datum = rawdatum + (qS * eMapVal) + (abs(qS) * dMapVal)
		datum = rawdatum + (abs(qS) * dMapVal)
		Slist.append(datum)
	else: Slist.append(1000000.)
	# calculate which atom is minimum value and load in list
	# atomtypes: NA = 7, S=6, A=5, O=4, N=3, C=2, H=1
	minvalue = min([Slist[i],Alist[i],Clist[i],Hlist[i],Nlist[i],NAlist[i],Olist[i]])
	if minvalue == NAlist[i]:
		atomtype = 7
	if minvalue == Slist[i]:
		atomtype = 6
	if minvalue == Alist[i]:
		atomtype = 5
	if minvalue == Nlist[i]:
		atomtype = 3
	if minvalue == Clist[i]:
		atomtype = 2
	if minvalue == Hlist[i]:
		atomtype = 1
	if minvalue == Olist[i]:
		atomtype = 4
	MINlist.append(minvalue)
	MINatom.append(atomtype)
	i = i + 1
# now load up a 3D array with zeros and fill with best affinity values
# load a second array with the best atom type
bestmap = zeros((nx,ny,nz),'float')
bestatom = zeros((nx,ny,nz), 'int')
listinc = 0
for k in range(nz):
	for j in range(ny):
		for i in range(nx):
			bestmap[i,j,k] = MINlist[listinc]
			bestatom[i,j,k] = MINatom[listinc]
			listinc = listinc + 1
# define function for finding neighbors
def nborfinder(x,nx,y,ny,z,nz,bestmap,bestatom,nbor,flood):
	"This funtion finds orthogonal neighbor points"
	if x+1 <= nx-1:
		node = [bestmap[x+1,y,z],x+1,y,z,bestatom[x+1,y,z]]
		if node not in nbor and node not in flood:
			nbor.append(node)
	if x-1 >= 0:
		node = [bestmap[x-1,y,z],x-1,y,z,bestatom[x-1,y,z]]
		if node not in nbor and node not in flood:
			nbor.append(node)
	if y+1 <= ny-1:
		node = [bestmap[x,y+1,z],x,y+1,z,bestatom[x,y+1,z]]
		if node not in nbor and node not in flood:
			nbor.append(node)
	if y-1 >= 0:
		node = [bestmap[x,y-1,z],x,y-1,z,bestatom[x,y-1,z]]
		if node not in nbor and node not in flood:
			nbor.append(node)
	if z+1 <= nz-1:
		node = [bestmap[x,y,z+1],x,y,z+1,bestatom[x,y,z+1]]
		if node not in nbor and node not in flood:
			nbor.append(node)
	if z-1 >= 0:
		node = [bestmap[x,y,z-1],x,y,z-1,bestatom[x,y,z-1]]
		if node not in nbor and node not in flood:
			nbor.append(node)
# define second function for finding neighbors with no steric hindrence
def nborfinder2(x,nx,y,ny,z,nz,bestmap,bestatom,nbor,flood):
	"This funtion finds orthogonal neighbor points with values <= 0"
	if x+1 <= nx-1:
		node = [bestmap[x+1,y,z],x+1,y,z,bestatom[x+1,y,z]]
		if node[0] <= 0.0:
			if node not in nbor and node not in flood:
				nbor.append(node)
	if x-1 >= 0:
		node = [bestmap[x-1,y,z],x-1,y,z,bestatom[x-1,y,z]]
		if node[0] <= 0.0:
			if node not in nbor and node not in flood:
				nbor.append(node)
	if y+1 <= ny-1:
		node = [bestmap[x,y+1,z],x,y+1,z,bestatom[x,y+1,z]]
		if node[0] <= 0.0:
			if node not in nbor and node not in flood:
				nbor.append(node)
	if y-1 >= 0:
		node = [bestmap[x,y-1,z],x,y-1,z,bestatom[x,y-1,z]]
		if node[0] <= 0.0:
			if node not in nbor and node not in flood:
				nbor.append(node)
	if z+1 <= nz-1:
		node = [bestmap[x,y,z+1],x,y,z+1,bestatom[x,y,z+1]]
		if node[0] <= 0.0:
			if node not in nbor and node not in flood:
				nbor.append(node)
	if z-1 >= 0:
		node = [bestmap[x,y,z-1],x,y,z-1,bestatom[x,y,z-1]]
		if node[0] <= 0.0:
			if node not in nbor and node not in flood:
				nbor.append(node)
# define function to set state for contiguous
def setstate(worstpoint,flood):
	"This function determines if a point is removable or not"
	# return a value for contig, 0 for removable and 1 for non-removable
	# create a dictionary of coords in flood and assign value to 1
	floodcoords = {}
	for fl in flood:
		floodcoords[fl[1],fl[2],fl[3]] = 1
	F = [0,0,0,0,0,0]
	tag = [0,0,0,0,0,0]
	nborcoords = []
	neighbors = 0
 	contig = 0  # defult is removable
	planes = 0
	x = worstpoint[1]
	y = worstpoint[2]
	z = worstpoint[3]
	# check for orthogonal neighbors and set F to 1 if there
	if floodcoords.has_key((x+1,y,z)):
		F[0] = 1
		neighbors = neighbors + 1
		nborcoords.append((x+1,y,z))
	if floodcoords.has_key((x-1,y,z)):
		F[3] = 1
		neighbors = neighbors + 1
		nborcoords.append((x-1,y,z))
	if floodcoords.has_key((x,y+1,z)):
		F[1] = 1
		neighbors = neighbors + 1
		nborcoords.append((x,y+1,z))
	if floodcoords.has_key((x,y-1,z)):
		F[4] = 1
		neighbors = neighbors + 1
		nborcoords.append((x,y-1,z))
	if floodcoords.has_key((x,y,z+1)):
		F[2] = 1
		neighbors = neighbors + 1
		nborcoords.append((x,y,z+1))
	if floodcoords.has_key((x,y,z-1)):
		F[5] = 1
		neighbors = neighbors + 1
		nborcoords.append((x,y,z-1))
	# check for planes, planes must tag each neighbor or point non-removable
	if floodcoords.has_key((x,y-1,z+1)) and F[4] == 1 and F[2] == 1:
		planes = planes + 1
		tag[4] = 1
		tag[2] = 1
	if floodcoords.has_key((x,y+1,z+1)) and F[1] == 1 and F[2] == 1:
		planes = planes + 1
		tag[1] = 1
		tag[2] = 1
	if floodcoords.has_key((x,y-1,z-1)) and F[4] == 1 and F[5] == 1:
		planes = planes + 1
		tag[4] = 1
		tag[5] = 1
	if floodcoords.has_key((x,y+1,z-1)) and F[1] == 1 and F[5] == 1:
		planes = planes + 1
		tag[1] = 1
		tag[5] = 1
	if floodcoords.has_key((x+1,y,z+1)) and F[2] == 1 and F[0] == 1:
		planes = planes + 1
		tag[2] = 1
		tag[0] = 1
	if floodcoords.has_key((x-1,y,z+1)) and F[2] == 1 and F[3] == 1:
		planes = planes + 1
		tag[2] = 1
		tag[3] = 1
	if floodcoords.has_key((x+1,y,z-1)) and F[0] == 1 and F[5] == 1:
		planes = planes + 1
		tag[0] = 1
		tag[5] = 1
	if floodcoords.has_key((x-1,y,z-1)) and F[3] == 1 and F[5] == 1:
		planes = planes + 1
		tag[3] = 1
		tag[5] = 1
	if floodcoords.has_key((x+1,y-1,z)) and F[0] == 1 and F[4] == 1:
		planes = planes + 1
		tag[0] = 1
		tag[4] = 1
	if floodcoords.has_key((x+1,y+1,z)) and F[0] == 1 and F[1] == 1:
		planes = planes + 1
		tag[0] = 1
		tag[1] = 1
	if floodcoords.has_key((x-1,y-1,z)) and F[3] == 1 and F[4] == 1:
		planes = planes + 1
		tag[3] = 1
		tag[4] = 1
	if floodcoords.has_key((x-1,y+1,z)) and F[1] == 1 and F[3] == 1:
		planes = planes + 1
		tag[1] = 1
		tag[3] = 1
	# now check for removable point and set state flag
	tagtotal = tag[0] + tag[1] + tag[2] + tag[3] + tag[4] + tag[5]
	if neighbors == 1:
		contig = 0  # state is set to removable
	if neighbors == 2:
		if planes >= 1 and tagtotal == 2:
			contig = 0
		else: contig = 1
	if neighbors == 3:
		if planes >= 2 and tagtotal == 3:
			contig = 0
		else: contig = 1
	if neighbors == 4:
		if planes >= 3 and tagtotal == 4:
			contig = 0
		else: contig = 1
	if neighbors == 5:
		if planes >= 4 and tagtotal == 5:
			contig = 0
		else: contig = 1
	if neighbors == 6:
		if planes >= 5 and tagtotal == 6:
			contig = 0
		else: contig = 1
	# now check if non-removable point is in a circuit and can be removed
	if contig == 1:
		# note, this can only be true if 2 or more neighbors
		# load up a test coordinate dictionary
		testfloodcoords = {}
		for tfl in flood:
			testfloodcoords[tfl[1],tfl[2],tfl[3]] = 1
		# start with the last coord added to nborcoords
		coord = nborcoords.pop()
		del testfloodcoords[(x,y,z)]  # remove starting point
		# load que with neighbors of the active point: ap
		Q = []
		apx = coord[0]
		apy = coord[1]
		apz = coord[2]
		if testfloodcoords.has_key((apx+1,apy,apz)):
			d = sqrt(((x-(apx+1))**2)+((y-apy)**2)+((z-apz)**2))
			Q.append((d,apx+1,apy,apz))
		if testfloodcoords.has_key((apx-1,apy,apz)):
			d = sqrt(((x-(apx-1))**2)+((y-apy)**2)+((z-apz)**2))
			Q.append((d,apx-1,apy,apz))
		if testfloodcoords.has_key((apx,apy+1,apz)):
			d = sqrt(((x-apx)**2)+((y-(apy+1))**2)+((z-apz)**2))
			Q.append((d,apx,apy+1,apz))
		if testfloodcoords.has_key((apx,apy-1,apz)):
			d = sqrt(((x-apx)**2)+((y-(apy-1))**2)+((z-apz)**2))
			Q.append((d,apx,apy-1,apz))
		if testfloodcoords.has_key((apx,apy,apz+1)):
			d = sqrt(((x-apx)**2)+((y-apy)**2)+((z-(apz+1))**2))
			Q.append((d,apx,apy,apz+1))
		if testfloodcoords.has_key((apx,apy,apz-1)):
			d = sqrt(((x-apx)**2)+((y-apy)**2)+((z-(apz-1))**2))
			Q.append((d,apx,apy,apz-1))
		del testfloodcoords[(apx,apy,apz)]  # delete the active point
		# now traverse the que and check for nborcoords, remove points as you go
		while 1:
			if len(Q) == 0:
				break
			Q.sort()
			qcoord = Q.pop(0)
			apx = qcoord[1]
			apy = qcoord[2]
			apz = qcoord[3]
			del testfloodcoords[(apx,apy,apz)]  # delete the new active point
			# check to see if this point is the last untouched neighbor point
			if nborcoords.count((apx,apy,apz)) == 1:
				nborcoords.remove((apx,apy,apz))
			if len(nborcoords) == 0:
				contig = 0  # all neighbor points still conected, so removable
				break
			if testfloodcoords.has_key((apx+1,apy,apz)):
				d = sqrt(((x-(apx+1))**2)+((y-apy)**2)+((z-apz)**2))
				if Q.count((d,apx+1,apy,apz)) == 0:
					Q.append((d,apx+1,apy,apz))
			if testfloodcoords.has_key((apx-1,apy,apz)):
				d = sqrt(((x-(apx-1))**2)+((y-apy)**2)+((z-apz)**2))
				if Q.count((d,apx-1,apy,apz)) == 0:
					Q.append((d,apx-1,apy,apz))
			if testfloodcoords.has_key((apx,apy+1,apz)):
				d = sqrt(((x-apx)**2)+((y-(apy+1))**2)+((z-apz)**2))
				if Q.count((d,apx,apy+1,apz)) == 0:
					Q.append((d,apx,apy+1,apz))
			if testfloodcoords.has_key((apx,apy-1,apz)):
				d = sqrt(((x-apx)**2)+((y-(apy-1))**2)+((z-apz)**2))
				if Q.count((d,apx,apy-1,apz)) == 0:
					Q.append((d,apx,apy-1,apz))
			if testfloodcoords.has_key((apx,apy,apz+1)):
				d = sqrt(((x-apx)**2)+((y-apy)**2)+((z-(apz+1))**2))
				if Q.count((d,apx,apy,apz+1)) == 0:
					Q.append((d,apx,apy,apz+1))
			if testfloodcoords.has_key((apx,apy,apz-1)):
				d = sqrt(((x-apx)**2)+((y-apy)**2)+((z-(apz-1))**2))
				if Q.count((d,apx,apy,apz-1)) == 0:
					Q.append((d,apx,apy,apz-1))
	return contig
# find the center node
xcent = int(nx/2)
ycent = int(ny/2)
zcent = int(nz/2)
if start_point_flag == 0:
	print 'done loading affinity maps, now track through all points and find (ten) best start points'
if start_point_flag == 2:
	print 'done loading affinity maps, now generate fill that contains both input points'
# first, outside mesh loop, set up a coord mods list to find 1A sphere around flood points for vol calc
CoordModsList = []
cpr = 1.0/spacing  # cpr = coord point radii
vpr = int(cpr)  # vpr = volume point radii - the number of gridpoints in radius
for vz in range(-vpr,vpr+1):
	for vy in range(-vpr,vpr+1):
		for vx in range(-vpr,vpr+1):
			voldist = sqrt((vx**2)+(vy**2)+(vz**2))
			if voldist <= cpr:
				CoordModsList.append([vx,vy,vz])
# also, make a 6 A sphere for the PP migration
# NOTE: if a 1A spacing is not used, then the PP method will search a smaller radii
CoordModsList6A = []
cpr = 6.0/spacing  # cpr = coord point radii
vpr = int(cpr)  # vpr = volume point radii - the number of gridpoints in radius
for vz in range(-vpr,vpr+1):
	for vy in range(-vpr,vpr+1):
		for vx in range(-vpr,vpr+1):
			voldist = sqrt((vx**2)+(vy**2)+(vz**2))
			if voldist <= cpr:
				CoordModsList6A.append([vx,vy,vz])
if start_point_flag == 0:
	mesh = []
	fullgrid = []
	fullmesh = []
	nonoverlapfills = []
	for k in range(0,nz,every):
		for j in range(0,ny,every):
			for i in range(0,nx,every):
				best = [i,j,k]
				fullgrid.append(best)
	# generate coords in cube about point in fullgrid
	StartPointModList = []
	for k in range(-1,2):
		for j in range(-1,2):
			for i in range(-1,2):
				StartPointModList.append([i,j,k])
	for fg in fullgrid:
		Shell = []
		for SPML in StartPointModList:
			x = fg[0]+SPML[0]
			if x >= nx or x < 0:
				continue
			y = fg[1]+SPML[1]
			if y >= ny or y < 0:
				continue
			z = fg[2]+SPML[2]
			if z >= nz or z < 0:
				continue
			node = [bestmap[x,y,z],x,y,z,bestatom[x,y,z]]
			if node[0] >= 0.0: # don't start inside receptor or far away
				continue
			if node[4] == 1:   # don't start with H atoms
				continue
			Shell.append(node)
		if len(Shell) == 0:
			continue
		# pick best point in shell
		Shell.sort()
		bestShell = Shell[0]
		spx = bestShell[1]
		spy = bestShell[2]
		spz = bestShell[3]
		# the start point should now be:  [spx,spy,spz]
		fill = []
		startpoint = [bestmap[spx,spy,spz],spx,spy,spz,bestatom[spx,spy,spz]]
		fill.append(startpoint)
		# now generate neighbors - save in growing list of all neighbors
		nbor = []
		TotalEnergy = startpoint[0]
		for i in range((volume - 1)):
			nborfinder(spx,nx,spy,ny,spz,nz,bestmap,bestatom,nbor,fill)
			# now sort neighbors list for min energy and add to fill list
			nbor.sort()
			node = nbor.pop(0)
			fill.append(node)
			TotalEnergy = TotalEnergy + node[0]
			spx = node[1]
			spy = node[2]
			spz = node[3]
		# now calculate volume of fill with 1A radii
		TotalVolume = []
		for fl in fill:
			for CML in CoordModsList:
				point = [fl[1]+CML[0],fl[2]+CML[1],fl[3]+CML[2]]
				if point not in TotalVolume:
					TotalVolume.append(point)
		NumberVolumePoints = len(TotalVolume)
		volangstrom = NumberVolumePoints * (spacing**3)
		epv = TotalEnergy/volangstrom
		fillpoint = [epv,startpoint[1],startpoint[2],startpoint[3],fill]
		fullmesh.append(fillpoint)
	fullmesh.sort()
	# now check if fills overlap and remove those that overlap by 80 percent
	# load the first fill into nonoverlapfills
	fp = fullmesh[0][4]
	for point in fp:
		nonoverlapfills.append(point)
	eightypercent = int(volume *0.8)
	tempmesh = fullmesh[1:]
	for fm in tempmesh:
		fp = fm[4]
		overlapcount = 0
		for point in fp:
			if point in nonoverlapfills:
				overlapcount = overlapcount + 1
		if overlapcount >= eightypercent:
			# remove fm from fullmesh
			fullmesh.remove(fm)
		else:
			for point in fp:
				nonoverlapfills.append(point)
	fmrange = len(fullmesh)
	if fmrange > Number_fills:
		fmrange = Number_fills
	for fm in range(fmrange):
		mesh.append(fullmesh[fm])
		testpoint = fullmesh[fm]
		x = (testpoint[1] - xcent)*spacing + centerx
		y = (testpoint[2] - ycent)*spacing + centery
		z = (testpoint[3] - zcent)*spacing + centerz
if start_point_flag == 1:
	# generate a startpoint from the input xyz
	mesh = []
	testmesh = []
	StartPointModList = []
	for k in range(-4,5):
		for j in range(-4,5):
			for i in range(-4,5):
				dist = sqrt((i**2)+(j**2)+(k**2))
				StartPointModList.append([dist,i,j,k])
	xpt = xcent + int((x_input - centerx)/spacing)
	ypt = ycent + int((y_input - centery)/spacing)
	zpt = zcent + int((z_input - centerz)/spacing)
	for SPML in StartPointModList:
		if xpt+SPML[1] >= nx-1 or xpt+SPML[1] <= 0:
			continue
		if ypt+SPML[2] >= ny-1 or ypt+SPML[2] <= 0:
			continue
		if zpt+SPML[3] >= nz-1 or zpt+SPML[3] <= 0:
			continue
		if bestmap[xpt+SPML[1],ypt+SPML[2],zpt+SPML[3]] >= 0.0:
			continue
		if bestatom[xpt+SPML[1],ypt+SPML[2],zpt+SPML[3]] == 1:
			continue
		fillpoint = [SPML[0],xpt+SPML[1],ypt+SPML[2],zpt+SPML[3]]
		testmesh.append(fillpoint)
	testmesh.sort()
	if len(testmesh) == 0:
		print 'no good starting points within 4A of your chosen point'
		sys.exit(2)
	goodfillpoint = testmesh[0]
	mesh.append(goodfillpoint)
if start_point_flag == 2:
	# fill mesh with one dummy point so that the meshpoint loop only runs once
	mesh = []
	dummyfillpoint = [0.,0,0,0]
	mesh.append(dummyfillpoint)
# now start with the startpoints in mesh
itmesh = 0
unusablepointcount = 0
badHstartFlag = 0
PDBlist = []
OUTlist = []
OUTcount = 0
for meshpoint in mesh:
	itmesh = itmesh + 1
	if start_point_flag == 2:
		# load in the two fixed start points and make sure they are valid
		StartPointModList = []
		for k in range(-4,5):
			for j in range(-4,5):
				for i in range(-4,5):
					dist = sqrt((i**2)+(j**2)+(k**2))
					StartPointModList.append([dist,i,j,k])
		# generate the first startpoint from the input xyz
		testmesh = []
		xpt = xcent + int((x_input - centerx)/spacing)
		ypt = ycent + int((y_input - centery)/spacing)
		zpt = zcent + int((z_input - centerz)/spacing)
		for SPML in StartPointModList:
			if xpt+SPML[1] >= nx-1 or xpt+SPML[1] <= 0:
				continue
			if ypt+SPML[2] >= ny-1 or ypt+SPML[2] <= 0:
				continue
			if zpt+SPML[3] >= nz-1 or zpt+SPML[3] <= 0:
				continue
			if bestmap[xpt+SPML[1],ypt+SPML[2],zpt+SPML[3]] > 0.0:
				continue
			fillpoint = [SPML[0],xpt+SPML[1],ypt+SPML[2],zpt+SPML[3]]
			testmesh.append(fillpoint)
		testmesh.sort()
		if len(testmesh) == 0:
			print 'no good starting points within 4A of 1st input point'
			sys.exit(2)
		gpf1 = testmesh[0] # good fill point coords
		loadpoint1 = [bestmap[gpf1[1],gpf1[2],gpf1[3]],gpf1[1],gpf1[2],gpf1[3],bestatom[gpf1[1],gpf1[2],gpf1[3]]]
		print 'first user point loaded'
		# generate the second startpoint from the input xyz
		testmesh = []
		xpt = xcent + int((x2_input - centerx)/spacing)
		ypt = ycent + int((y2_input - centery)/spacing)
		zpt = zcent + int((z2_input - centerz)/spacing)
		for SPML in StartPointModList:
			if xpt+SPML[1] >= nx-1 or xpt+SPML[1] <= 0:
				continue
			if ypt+SPML[2] >= ny-1 or ypt+SPML[2] <= 0:
				continue
			if zpt+SPML[3] >= nz-1 or zpt+SPML[3] <= 0:
				continue
			if bestmap[xpt+SPML[1],ypt+SPML[2],zpt+SPML[3]] > 0.0:
				continue
			fillpoint = [SPML[0],xpt+SPML[1],ypt+SPML[2],zpt+SPML[3]]
			testmesh.append(fillpoint)
		testmesh.sort()
		if len(testmesh) == 0:
			print 'no good starting points within 4A of 2nd input point'
			sys.exit(2)
		gpf2 = testmesh[0] # good fill point coords
		loadpoint2 = [bestmap[gpf2[1],gpf2[2],gpf2[3]],gpf2[1],gpf2[2],gpf2[3],bestatom[gpf2[1],gpf2[2],gpf2[3]]]
		print 'second user point loaded'
		# exclude steric overlap points (use only 0 and negative affinity)
		# add points from first to second along legal pathways - then bring fill up to input volume
		spx = loadpoint1[1]
		spy = loadpoint1[2]
		spz = loadpoint1[3]
		startingpoint = (spx,spy,spz)
		flood = []
		startpoint = [bestmap[spx,spy,spz],spx,spy,spz,bestatom[spx,spy,spz]]
		flood.append(startpoint)
		# now generate neighbors - save in growing list of all neighbors
		# add neighbor point to flood that gets closer to loadpoint2
		nbor = []
		reached_B = 0
		for i in range(volume-1):
			nborfinder2(spx,nx,spy,ny,spz,nz,bestmap,bestatom,nbor,flood)
			# now sort neighbors list for min distance and add to flood list
			if reached_B == 1:
				nbor.sort()
			else:
				nborD = []
				for n in nbor:
					DtoB = sqrt((n[1] - gpf2[1])**2.0 + (n[2] - gpf2[2])**2.0 + (n[3] - gpf2[3])**2.0)
					if DtoB == 0.0:
						reached_B = 1
					nborD.append([DtoB,n[1],n[2],n[3]])
				# sort nborD and reload nbor in that order
				nborD.sort()
				nbor = []
				for n in nborD:
					node = [bestmap[n[1],n[2],n[3]],n[1],n[2],n[3],bestatom[n[1],n[2],n[3]]]
					nbor.append(node)
			# node = nbor.pop(0)
			node = nbor[0]
			Hloopcount = 0
			nborsize = len(nbor)
			while node[4] == 1:  # H atom
				x = node[1]
				y = node[2]
				z = node[3]
				Hstable = 0
				if x+1 <= nx-1:
					Hnode = [bestmap[x+1,y,z],x+1,y,z,bestatom[x+1,y,z]]
					if Hnode in flood and bestatom[x+1,y,z] != 1:
						Hstable = 1
				if x-1 >= 0:
					Hnode = [bestmap[x-1,y,z],x-1,y,z,bestatom[x-1,y,z]]
					if Hnode in flood and bestatom[x-1,y,z] != 1:
						Hstable = 1
				if y+1 <= ny-1:
					Hnode = [bestmap[x,y+1,z],x,y+1,z,bestatom[x,y+1,z]]
					if Hnode in flood and bestatom[x,y+1,z] != 1:
						Hstable = 1
				if y-1 >= 0:
					Hnode = [bestmap[x,y-1,z],x,y-1,z,bestatom[x,y-1,z]]
					if Hnode in flood and bestatom[x,y-1,z] != 1:
						Hstable = 1
				if z+1 <= nz-1:
					Hnode = [bestmap[x,y,z+1],x,y,z+1,bestatom[x,y,z+1]]
					if Hnode in flood and bestatom[x,y,z+1] != 1:
						Hstable = 1
				if z-1 >= 0:
					Hnode = [bestmap[x,y,z-1],x,y,z-1,bestatom[x,y,z-1]]
					if Hnode in flood and bestatom[x,y,z-1] != 1:
						Hstable = 1
				if Hstable == 0:
					Hloopcount = Hloopcount + 1
					if Hloopcount == nborsize:
						badHstartFlag = 1
						break
					node = nbor[Hloopcount]
					continue
				else:
					break
			if badHstartFlag == 1:
				break
			node = nbor.pop(Hloopcount)
			flood.append(node)
			spx = node[1]
			spy = node[2]
			spz = node[3]
		if badHstartFlag == 1:
			badHstartFlag = 0   # reset flag
			continue
		# need to finish neighbor nodes for last flood point
		nborfinder(spx,nx,spy,ny,spz,nz,bestmap,bestatom,nbor,flood)
		nbor.sort()
		# update nbor
		nbor = []
		TotalEnergy = 0.0
		for fl in flood:
			nborx = fl[1]
			nbory = fl[2]
			nborz = fl[3]
			nborfinder(nborx,nx,nbory,ny,nborz,nz,bestmap,bestatom,nbor,flood)
			TotalEnergy = TotalEnergy + fl[0]
	else:
		spx = meshpoint[1]
		spy = meshpoint[2]
		spz = meshpoint[3]
		if bestmap[spx,spy,spz] >= 0.0:
			unusablepointcount = unusablepointcount + 1
			continue
		# the start point should now be:  [spx,spy,spz]
		startingpoint = (spx,spy,spz)
		flood = []
		startpoint = [bestmap[spx,spy,spz],spx,spy,spz,bestatom[spx,spy,spz]]
		flood.append(startpoint)
		# now generate neighbors - save in growing list of all neighbors
		nbor = []
		TotalEnergy = startpoint[0]
		
		for i in range(volume-1):
			#this is where we need to build the display
			#flood is list of list x,y,z,flood[j][1:4] flood[j][4] = atom type
			# GUI
			if movie:
				save_flood(flood)
			nborfinder(spx,nx,spy,ny,spz,nz,bestmap,bestatom,nbor,flood)
			# now sort neighbors list for min energy and add to flood list
			nbor.sort()
			node = nbor[0]
			Hloopcount = 0
			nborsize = len(nbor)
			while node[4] == 1:  # H atom
				x = node[1]
				y = node[2]
				z = node[3]
				Hstable = 0
				if x+1 <= nx-1:
					Hnode = [bestmap[x+1,y,z],x+1,y,z,bestatom[x+1,y,z]]
					if Hnode in flood and bestatom[x+1,y,z] != 1:
						Hstable = 1
				if x-1 >= 0:
					Hnode = [bestmap[x-1,y,z],x-1,y,z,bestatom[x-1,y,z]]
					if Hnode in flood and bestatom[x-1,y,z] != 1:
						Hstable = 1
				if y+1 <= ny-1:
					Hnode = [bestmap[x,y+1,z],x,y+1,z,bestatom[x,y+1,z]]
					if Hnode in flood and bestatom[x,y+1,z] != 1:
						Hstable = 1
				if y-1 >= 0:
					Hnode = [bestmap[x,y-1,z],x,y-1,z,bestatom[x,y-1,z]]
					if Hnode in flood and bestatom[x,y-1,z] != 1:
						Hstable = 1
				if z+1 <= nz-1:
					Hnode = [bestmap[x,y,z+1],x,y,z+1,bestatom[x,y,z+1]]
					if Hnode in flood and bestatom[x,y,z+1] != 1:
						Hstable = 1
				if z-1 >= 0:
					Hnode = [bestmap[x,y,z-1],x,y,z-1,bestatom[x,y,z-1]]
					if Hnode in flood and bestatom[x,y,z-1] != 1:
						Hstable = 1
				if Hstable == 0:
					Hloopcount = Hloopcount + 1
					if Hloopcount == nborsize:
						badHstartFlag = 1
						break
					node = nbor[Hloopcount]
					continue
				else:
					break
			if badHstartFlag == 1:
				break
			node = nbor.pop(Hloopcount)
			flood.append(node)
			spx = node[1]
			spy = node[2]
			spz = node[3]
			TotalEnergy = TotalEnergy + node[0]
		if badHstartFlag == 1:
			badHstartFlag = 0   # reset flag
			continue
		# need to finish neighbor nodes for last flood point
		nborfinder(spx,nx,spy,ny,spz,nz,bestmap,bestatom,nbor,flood)
		nbor.sort()
	# now let the flooded area migrate to higher affinity area
	flagPP = 1
	PPcount = 0
	while flagPP == 1:
		# GUI
		if movie:
			save_flood(flood)
		flagPP = 0
		flood.sort()
		worstpoint = flood[volume-1]
		bestpoint = nbor[0]
		loopcount = 0
		Hloopcount = 0
		EPVloopcount = 0
		# find the energy per volume for initial fill
		TotalVolume = []
		for fl in flood:
			for CML in CoordModsList:
				point = [fl[1]+CML[0],fl[2]+CML[1],fl[3]+CML[2]]
				if point not in TotalVolume:
					TotalVolume.append(point)
		NumberVolumePoints = len(TotalVolume)
		volangstrom = NumberVolumePoints * (spacing**3)
		ThisEPV = TotalEnergy/volangstrom
		while bestpoint[0] < worstpoint[0]:
			# GUI
			if movie:
				save_flood(flood)			

			# check to see if removing worstpoint from flood breaks continuity
			#   look at state varible contig: 0 = removable, 1 = non-removable
 			contig = setstate(worstpoint,flood)
			if start_point_flag == 2:
				if worstpoint == loadpoint1:
					contig = 1
				if worstpoint == loadpoint2:
					contig = 1
			# if still contiguous: remove worstpoint and add bestpoint to flood
			# then recalculate nbor shell
			# but not if bestpoint is H and only neighbor is H atom
			if bestpoint[4] == 1 and contig == 0:
				x = bestpoint[1]
				y = bestpoint[2]
				z = bestpoint[3]
				Hstable = 0
				if x+1 <= nx-1:
					node = [bestmap[x+1,y,z],x+1,y,z,bestatom[x+1,y,z]]
					if node in flood and bestatom[x+1,y,z] != 1:
						Hstable = 1
				if x-1 >= 0:
					node = [bestmap[x-1,y,z],x-1,y,z,bestatom[x-1,y,z]]
					if node in flood and bestatom[x-1,y,z] != 1:
						Hstable = 1
				if y+1 <= ny-1:
					node = [bestmap[x,y+1,z],x,y+1,z,bestatom[x,y+1,z]]
					if node in flood and bestatom[x,y+1,z] != 1:
						Hstable = 1
				if y-1 >= 0:
					node = [bestmap[x,y-1,z],x,y-1,z,bestatom[x,y-1,z]]
					if node in flood and bestatom[x,y-1,z] != 1:
						Hstable = 1
				if z+1 <= nz-1:
					node = [bestmap[x,y,z+1],x,y,z+1,bestatom[x,y,z+1]]
					if node in flood and bestatom[x,y,z+1] != 1:
						Hstable = 1
				if z-1 >= 0:
					node = [bestmap[x,y,z-1],x,y,z-1,bestatom[x,y,z-1]]
					if node in flood and bestatom[x,y,z-1] != 1:
						Hstable = 1
				if Hstable == 0:
					Hloopcount = Hloopcount + 1
					bestpoint = nbor[Hloopcount]
					continue
			if contig == 0:
				# first check if bestpoint is only next to worstpoint
				x = bestpoint[1]
				y = bestpoint[2]
				z = bestpoint[3]
				Stable = 0
				if x+1 <= nx-1:
					node = [bestmap[x+1,y,z],x+1,y,z,bestatom[x+1,y,z]]
					if node in flood:
						Stable += 1
					if node == worstpoint:
						Stable -= 1
				if x-1 >= 0:
					node = [bestmap[x-1,y,z],x-1,y,z,bestatom[x-1,y,z]]
					if node in flood:
						Stable += 1
					if node == worstpoint:
						Stable -= 1
				if y+1 <= ny-1:
					node = [bestmap[x,y+1,z],x,y+1,z,bestatom[x,y+1,z]]
					if node in flood:
						Stable += 1
					if node == worstpoint:
						Stable -= 1
				if y-1 >= 0:
					node = [bestmap[x,y-1,z],x,y-1,z,bestatom[x,y-1,z]]
					if node in flood:
						Stable += 1
					if node == worstpoint:
						Stable -= 1
				if z+1 <= nz-1:
					node = [bestmap[x,y,z+1],x,y,z+1,bestatom[x,y,z+1]]
					if node in flood:
						Stable += 1
					if node == worstpoint:
						Stable -= 1
				if z-1 >= 0:
					node = [bestmap[x,y,z-1],x,y,z-1,bestatom[x,y,z-1]]
					if node in flood:
						Stable += 1
					if node == worstpoint:
						Stable -= 1
				if Stable <= 0:
					loopcount = loopcount + 1
					if (volume - loopcount) < 1:
						break
					worstpoint = flood[volume -1 -loopcount]
					continue
				# check if epv is beter with worstpoint changed to bestpoint
				# if not, use next best nbor point
				TotalVolume = []
				for fl in flood:
					for CML in CoordModsList:
						point = [fl[1]+CML[0],fl[2]+CML[1],fl[3]+CML[2]]
						if point not in TotalVolume:
							TotalVolume.append(point)
				NumberVolumePoints = len(TotalVolume)
				volangstrom = NumberVolumePoints * (spacing**3)
				epv = TotalEnergy/volangstrom
				if epv > ThisEPV:
					EPVloopcount = EPVloopcount + 1
					bestpoint = nbor[EPVloopcount]
					continue
				# now add bestpoint and remove worstpoint
				flood.pop(volume-1-loopcount)
				flood.append(bestpoint)
				flood.sort()
				# update nbor
				nbor = []
				for fl in flood:
					nborx = fl[1]
					nbory = fl[2]
					nborz = fl[3]
					nborfinder(nborx,nx,nbory,ny,nborz,nz,bestmap,bestatom,nbor,flood)
				nbor.sort()
				worstpoint = flood[volume-1]
				bestpoint = nbor[0]
				loopcount = 0  # reset every time a point is moved
				Hloopcount = 0 # reset every time a point is moved
			else:
				loopcount = loopcount + 1
				if (volume - loopcount) < 1:
					break
				worstpoint = flood[volume -1 -loopcount]
		# Psuedo Pod method - reach out 6A away and look for beter energy
		# only run Psuedo Pod method on fills > 11 points
		if volume <= 11:
			continue
		# only let the psuedo pods run 100 times
		if PPcount > 100:
			print "Psuedo Pod over 100 - continuing..."
			continue
		PPcount = PPcount + 1
		# first, add up energy of 10 worst points
		flood.sort()
		wplist = []
		wpvalue = []
		wp = 0
		wpcount = 0
		# create temp flood list and remove worst points so that contig will not use removed points
		tempwpflood = []
		for fl in flood:
			tempwpflood.append(fl)
		while wpcount < volume:
			worstpoint = flood[volume -1 -wpcount]
			contig = setstate(worstpoint,tempwpflood)
			if start_point_flag == 2:
				if worstpoint == loadpoint1:
					contig = 1
				if worstpoint == loadpoint2:
					contig = 1
			if contig == 0:
				tempwpflood.remove(worstpoint)
				if wp == 0:
					wp1 = worstpoint[0]
					wpvalue.append(wp1)
					wplist.append(worstpoint)
				if wp == 1:
					wp2 = worstpoint[0] + wp1
					wpvalue.append(wp2)
					wplist.append(worstpoint)
				if wp == 2:
					wp3 = worstpoint[0] + wp2
					wpvalue.append(wp3)
					wplist.append(worstpoint)
				if wp == 3:
					wp4 = worstpoint[0] + wp3
					wpvalue.append(wp4)
					wplist.append(worstpoint)
				if wp == 4:
					wp5 = worstpoint[0] + wp4
					wpvalue.append(wp5)
					wplist.append(worstpoint)
				if wp == 5:
					wp6 = worstpoint[0] + wp5
					wpvalue.append(wp6)
					wplist.append(worstpoint)
				if wp == 6:
					wp7 = worstpoint[0] + wp6
					wpvalue.append(wp7)
					wplist.append(worstpoint)
				if wp == 7:
					wp8 = worstpoint[0] + wp7
					wpvalue.append(wp8)
					wplist.append(worstpoint)
				if wp == 8:
					wp9 = worstpoint[0] + wp8
					wpvalue.append(wp9)
					wplist.append(worstpoint)
				if wp == 9:
					wp10 = worstpoint[0] + wp9
					wpvalue.append(wp10)
					wplist.append(worstpoint)
					wp = 10
					break
				wp = wp + 1
			wpcount = wpcount + 1
		# next, loop through flood points and project pseudopods out to 6A
		# but not from the worst 10 nodes or from H atoms
		bestPP = []
		for fl in flood:
			if fl in wplist:
				continue
			if fl[4] == 1:   # atomtype 1 is an H atom
				continue
			totenergyPP = 0.0
			lengthPP = 0
			PPlist = []
			# generate 6A shell around point, exclude points in protein and flood
			Shell = []
			for CML6 in CoordModsList6A:
				x = fl[1]+CML6[0]
				if x >= nx or x < 0:
					continue
				y = fl[2]+CML6[1]
				if y >= ny or y < 0:
					continue
				z = fl[3]+CML6[2]
				if z >= nz or z < 0:
					continue
				node = [bestmap[x,y,z],x,y,z,bestatom[x,y,z]]
				if node[0] > 0.0:
					continue
				if node in flood:
					continue
				Shell.append(node)
			# pick best point in shell
			Shell.sort()
			bestShellpoint = Shell[0]
			# march from flood point to best Shell point - record length and total energy
			currentPPnode = [fl[0],fl[1],fl[2],fl[3],fl[4]]
			bspx = bestShellpoint[1]
			bspy = bestShellpoint[2]
			bspz = bestShellpoint[3]
			for i in range(10):
				x = currentPPnode[1]
				y = currentPPnode[2]
				z = currentPPnode[3]
				# check each direction from current point to get closer to best point
				checkNode = []
				subchecknode = []
				if x+1 <= nx-1:
					nodePP = [bestmap[x+1,y,z],x+1,y,z,bestatom[x+1,y,z]]
					if nodePP in Shell:
						DtoBSPsq = (nodePP[1] - bspx)**2.0 + (nodePP[2] -bspy)**2.0 + (nodePP[3] - bspz)**2.0
						subchecknode = [DtoBSPsq,nodePP]
						checkNode.append(subchecknode)
				if x-1 >= 0:
					nodePP = [bestmap[x-1,y,z],x-1,y,z,bestatom[x-1,y,z]]
					if nodePP in Shell:
						DtoBSPsq = (nodePP[1] - bspx)**2.0 + (nodePP[2] -bspy)**2.0 + (nodePP[3] - bspz)**2.0
						subchecknode = [DtoBSPsq,nodePP]
						checkNode.append(subchecknode)
				if y+1 <= ny-1:
					nodePP = [bestmap[x,y+1,z],x,y+1,z,bestatom[x,y+1,z]]
					if nodePP in Shell:
						DtoBSPsq = (nodePP[1] - bspx)**2.0 + (nodePP[2] -bspy)**2.0 + (nodePP[3] - bspz)**2.0
						subchecknode = [DtoBSPsq,nodePP]
						checkNode.append(subchecknode)
				if y-1 >= 0:
					nodePP = [bestmap[x,y-1,z],x,y-1,z,bestatom[x,y-1,z]]
					if nodePP in Shell:
						DtoBSPsq = (nodePP[1] - bspx)**2.0 + (nodePP[2] -bspy)**2.0 + (nodePP[3] - bspz)**2.0
						subchecknode = [DtoBSPsq,nodePP]
						checkNode.append(subchecknode)
				if z+1 <= nz-1:
					nodePP = [bestmap[x,y,z+1],x,y,z+1,bestatom[x,y,z+1]]
					if nodePP in Shell:
						DtoBSPsq = (nodePP[1] - bspx)**2.0 + (nodePP[2] -bspy)**2.0 + (nodePP[3] - bspz)**2.0
						subchecknode = [DtoBSPsq,nodePP]
						checkNode.append(subchecknode)
				if z-1 >= 0:
					nodePP = [bestmap[x,y,z-1],x,y,z-1,bestatom[x,y,z-1]]
					if nodePP in Shell:
						DtoBSPsq = (nodePP[1] - bspx)**2.0 + (nodePP[2] -bspy)**2.0 + (nodePP[3] - bspz)**2.0
						subchecknode = [DtoBSPsq,nodePP]
						checkNode.append(subchecknode)
				if len(checkNode) == 0:
					break
				lengthPP = lengthPP + 1
				checkNode.sort()
				subchecknode = checkNode[0]
				nodePP = subchecknode[1]
				PPlist.append(nodePP)
				# remove nodePP from Shell
				Shell.remove(nodePP)
				totenergyPP = totenergyPP + nodePP[0]
				if subchecknode[0] <= 0.01:
					# reached the best node
					break
				# stop at the first H reached
				if nodePP[4] == 1:
					break
				currentPPnode = [nodePP[0],nodePP[1],nodePP[2],nodePP[3],nodePP[4]]
			bestnodePP = [totenergyPP,PPlist,lengthPP]
			bestPP.append(bestnodePP)
		bestPP.sort()
		if len(bestPP) == 0:
			break
		bestnodePP = bestPP.pop(0)
		# if the pseudopod is better than up to 10 worst, remove worst and add pseudopod
		if bestnodePP[2] > len(wpvalue):
			break
		if bestnodePP[0] < wpvalue[bestnodePP[2]-1]:
			for i in range(bestnodePP[2]):
				flood.remove(wplist[i])
			for nodePP in bestnodePP[1]:
				flood.append(nodePP)
			# now update nbor and start over to migrate
			nbor = []
			for fl in flood:
				nborx = fl[1]
				nbory = fl[2]
				nborz = fl[3]
				nborfinder(nborx,nx,nbory,ny,nborz,nz,bestmap,bestatom,nbor,flood)
			nbor.sort()
			flagPP = 1 # repeat the migration loop to update with the pseudopod

	# now output flood as xyz file
	i = 0
	TotalEnergy = 0.0
	for fl in flood:
		TotalEnergy = TotalEnergy + fl[0]
	spconvx = (startingpoint[0] - xcent)*spacing + centerx
	spconvy = (startingpoint[1] - ycent)*spacing + centery
	spconvz = (startingpoint[2] - zcent)*spacing + centerz
	PDBsublist = []
	for fl in flood:
		i = i + 1
		x = (fl[1] - xcent)*spacing + centerx
		y = (fl[2] - ycent)*spacing + centery
		z = (fl[3] - zcent)*spacing + centerz
		if fl[4] == 7:
			atomchr = 'N'
		if fl[4] == 6:
			atomchr = 'S'
		if fl[4] == 5:
			atomchr = 'A'
		if fl[4] == 4:
			atomchr = 'O'
		if fl[4] == 3:
			atomchr = 'P'
			# note, this will color the N atom pink (the PDB color for Phosphorus)
		if fl[4] == 2:
			atomchr = 'C'
		if fl[4] == 1:
			atomchr = 'H'
		afin = abs(fl[0])
		PDBsublist.append([i,atomchr,x,y,z,afin])
	PDBlist.append(PDBsublist)
	# now calculate volume of flood with 1A radii
	TotalVolume = []
	for fl in flood:
		for CML in CoordModsList:
			point = [fl[1]+CML[0],fl[2]+CML[1],fl[3]+CML[2]]
			if point not in TotalVolume:
				TotalVolume.append(point)
	NumberVolumePoints = len(TotalVolume)
	volangstrom = NumberVolumePoints * (spacing**3)
	epv = TotalEnergy/volangstrom
	OUTcount = OUTcount + 1
	OUTlist.append([epv,volangstrom,OUTcount])

# GUI
if movie:
	save_flood(flood)
	output.close()

# sort and generate outputs
OUTlist.sort()
OUTlength = len(OUTlist)
lastOP = OUTlist[0]
# find non-duplicates
newOUTlist = []
newOUTlist.append(lastOP)
for i in range(OUTlength - 1):
	x = i + 1
	thisOP = OUTlist[x]
	if thisOP[0] == lastOP[0]:
		continue
	else:
		lastOP = thisOP
		newOUTlist.append(thisOP)
resultName = FileBaseName + '_' + str(volume) + 'Results.txt'
results_out = open(resultName,'w')
i = 0
if start_point_flag == 0:
	print '10 Unique Results:  (note, some starting points converge to same answer, so may be less than 10)'
for OP in newOUTlist:
	print 'Output #',i+1,' Total Volume = ',OP[1],' Total Energy per Vol, EPV = ',OP[0],' (Kcal/mol A**3)'
	results_out.write('Output #%3i, Total Volume = %7.2f, Total Energy per Vol, EPV = %15.8f\n' 
			   % (i+1,OP[1],OP[0]))
	# point loops
	PDBname = 'FILL_' + str(volume) + 'out' + str(i+1) + '.pdb'
	PDB_out = open(PDBname,'w')
	for PDB in PDBlist[OP[2]-1]:
		PDB_out.write('ATOM%7d  %s   FIL     1%12.3f%8.3f%8.3f  1.0%8.3f     0.001 %s\n' %
				(PDB[0],PDB[1],PDB[2],PDB[3],PDB[4],PDB[5],PDB[1]))
	PDB_out.close()
	i = i + 1
# clean house
results_out.close()
if Sflag == 1: Smap.close()
if Aflag == 1: Amap.close()
if Cflag == 1: Cmap.close()
if Hflag == 1: Hmap.close()
if Nflag == 1: Nmap.close()
if NAflag == 1: NAmap.close()
if Oflag == 1: Omap.close()
emap.close()
dmap.close()



#############################################################################
#
# Author: Ruth HUEY, Stefano FORLI
#
# Copyright: A. Olson TSRI 2010
#
#############################################################################


# $Header: /opt/cvs/python/packages/share1.5/AutoDockTools/InteractionDetector.py,v 1.17 2010/10/06 18:07:19 rhuey Exp $ 
#
# $Id: InteractionDetector.py,v 1.17 2010/10/06 18:07:19 rhuey Exp $
#
#
#
#
#
#
#

"""
Assorted classes for use in detecting interactions of various subgroups in AutoDock Virtual Screening results...

interactionDetector_instance.screen(pdbqt_result) returns interactions_found as True/False. 
If interactions_exist, a new pdbqt_result file including information describing the interactions is written.  
Types of interactions to be detected include:
    -contacts between receptor atoms and ligand atoms closer than sum of vdw radii
    -hydrogen-bond interactions between receptor atoms and ligand atoms of appropriate types
    -pi-pi interactions between receptor atoms and ligand atoms of appropriate types
    -cation-pi interactions between receptor atoms and ligand atoms of appropriate types
    -t_shaped interactions between receptor atoms and ligand atoms of appropriate types

The filter is initialized with a file containing the receptor residues whose interactions with ligands are of interest.
For example, in the case of hsg1, a file could potentially contain only A:ARG8 and B:ARG8 atoms.

"""
from string import strip
from numpy import oldnumeric as Numeric
from MolKit import Read
from MolKit.molecule import MoleculeSet, Atom, AtomSet, HydrogenBond, BondSet
from MolKit.protein import Residue, ResidueSet
from MolKit.distanceSelector import CloserThanVDWSelector, DistanceSelector
from MolKit.hydrogenBondBuilder import HydrogenBondBuilder
from PyBabel.cycle import RingFinder
from PyBabel.aromatic import Aromatic
from MolKit.bondSelector import AromaticCycleBondSelector
import numpy.oldnumeric as Numeric
from mglutil.math import crossProduct


class InteractionDetector:
    """
        Base class for object to detect interactions between a receptor and potential ligands... such as virtual screen results
        
        initialized with the receptor file containing entire molecule or residues of interest.
        processLigand method takes as input a pdbqt file containing docked coordinates and returns a string 
                        composed of  hbondStr + macrocloseContactStr + ligcloseContactStr
        If interections are found, a new pdbqt containing interaction description is also output: 
    """
    def __init__(self, receptor_file, percentCutoff=1., detect_pi=False, dist_cutoff=6, verbose=False, 
                        distanceSelector=None, hydrogen_bond_builder=None, distanceSelectorWithCutoff=None,
                        aromatic_cycle_bond_selector=None): 
        self.receptor_file = receptor_file
        receptor = Read(receptor_file)
        assert(len(receptor)==1)
        assert isinstance(receptor, MoleculeSet)
        self.macro = receptor[0]
        self.macro_atoms = self.macro.allAtoms
        self.macro.buildBondsByDistance()
        self.verbose = verbose
        #??useful??
        self.percentCutoff = percentCutoff
        self.detect_pi = detect_pi
        self.distanceSelector = distanceSelector 
        if self.distanceSelector is None:
            self.distanceSelector = CloserThanVDWSelector(return_dist=0)
        self.hydrogen_bond_builder = hydrogen_bond_builder 
        if self.hydrogen_bond_builder is None:
            self.hydrogen_bond_builder = HydrogenBondBuilder()
        self.distanceSelectorWithCutoff = distanceSelectorWithCutoff 
        if self.distanceSelectorWithCutoff is None:
            self.distanceSelectorWithCutoff = DistanceSelector()
        self.dist_cutoff=float(dist_cutoff)
        self.results = d = {}
        self.report_list =['lig_hb_atoms','lig_close_atoms']
        self.aromatic_cycle_bond_selector = aromatic_cycle_bond_selector
        if self.aromatic_cycle_bond_selector is None:
            self.aromatic_cycle_bond_selector = AromaticCycleBondSelector()
        if detect_pi:
            self.report_list.extend(['pi_cation','pi_pi', 'cation_pi', 't_shaped'])
        if self.verbose: print "self.report_list=", self.report_list


    def processLigand(self, ligand_file, percentCutoff=None, output=1, outputfilename=None, comment="USER  AD> ", buildHB=1, remove_modelStr=False):
        #if outputfilename is not None, have to (1) update all the strings or (2) rename ligand
        self.results = d = {}
        ligand = Read(ligand_file)
        assert isinstance(ligand, MoleculeSet)
        assert(len(ligand)==1)
        ligand = ligand[0]
        first = ligand.name.find('_model')
        last = ligand.name.rfind('_model')
        if first!=last:
            ligand.name = ligand.name[:last]
        if remove_modelStr:
            curName = ligand.name
            modelIndex=curName.find("_model")
            if modelIndex>-1:
                curName = curName[:modelIndex]
            #setup outputfilestem
            ligand.name = curName
        ligand.buildBondsByDistance()
        if not percentCutoff:
            percentCutoff = self.percentCutoff
        # first detect sets of atoms forming hydrogen bonds
        # hbondStr,hblist and has_hbonds added to process vina results which do not include valid hydrogen bonds
        hbondStr = ""
        hblist = [""]
        has_hbonds = False
        if buildHB:
            has_hbonds = True
            hbondStr = self.buildHydrogenBonds(ligand, comment=comment)           #
            if self.verbose: print "hbondStr=", hbondStr
            hblist = hbondStr.split('\n') # hbond info
            if hblist[0]=='lig_hb_atoms : 0':
                has_hbonds = False
        # next detect sets of atoms in close contact not forming hydrogen bonds
        macrocloseContactStr, ligcloseContactStr = self.buildCloseContactAtoms(percentCutoff, ligand, comment=comment)              #
        self.piResults = ""
        if self.detect_pi: 
            self.detectPiInteractions()
            if self.results['pi_cation']:
                self.piResults = self.get_pi_cation_result(print_ctr=1, comment=comment)
                if self.verbose: 
                    print "found pi_cation in ", ligand_file
                    print "set self.piResults to ", self.piResults
            if self.results['pi_pi']:
                self.piResults += self.get_pi_pi(print_ctr=1, comment=comment)
                if self.verbose: 
                    print "set self.piResults to ", self.piResults
        macro_cclist = macrocloseContactStr.split(';')
        lig_cclist = ligcloseContactStr.split(';')
        if has_hbonds or len(macro_cclist):
            fptr = open(ligand_file)
            lines = fptr.readlines()
            fptr.close()
            if outputfilename is not None:
                optr = open(outputfilename, 'w')
            else:
                optr = open(ligand_file, 'w')
            for new_l in hblist:
                if not(len(new_l)): #don't output empty lines
                    continue
                if new_l.find(comment)!=0:
                    new_l = comment + new_l
                if new_l != hblist[-1]:
                    optr.write(new_l+"\n")
                else:
                    optr.write(new_l) # no newline after the last one
            for new_l in macro_cclist:
                if not(len(new_l)): #don't output empty lines
                    continue
                if new_l.find(comment)!=0:
                    new_l = comment + new_l
                if new_l != macro_cclist[-1]:
                    optr.write(new_l + "\n")
                else:
                    optr.write(new_l)
            for new_l in lig_cclist:
                if not(len(new_l)): #don't output empty lines
                    continue
                if new_l.find(comment)!=0:
                    new_l = comment + new_l
                if new_l != lig_cclist[-1]:
                    optr.write(new_l + "\n")
                else:
                    optr.write(new_l)
            if len(self.piResults): ##???
                for s in self.piResults.split("\n"):
                    optr.write(s+"\n")
            for l in lines:
                optr.write(l)
            optr.close()
        return hbondStr + macrocloseContactStr + ligcloseContactStr
            

    def getResultStr(self, verbose=False):
        #only write important ones: hbonds, vdw_contacts, ?pi_pi,pi_cation etc?
        ss = ""
        for k in self.report_list:
            v = self.results[k]
            #key:atom1.full_name();
            if len(v):
                if verbose: print "gRS: report for ", k
                if verbose: print  "USER:  " + k + ":" + str(len(v)) 
                if verbose: print  "     start ss= ", ss
                ss +=  k + ":" + str(len(v))  + "\n"
                if k=='lig_hb_atoms':
                    if verbose: print "@@ getResultStr lig_hb_atoms"
                    hbstr = ""
                    for lig_at in v:
                        for hb in lig_at.hbonds:
                            if hasattr(hb, 'hAt'):
                                hbstr += 'USER %s-%s~%s\n'%(hb.donAt.full_name(), hb.hAt.full_name(),hb.accAt.full_name())
                            else:
                                hbstr += 'USER %s~%s\n'%(hb.donAt.full_name(), hb.accAt.full_name())
                    if verbose: print "hbstr="
                    #add it to ss here
                    ss += hbstr
                    if verbose: print "with hbstr: ss=", ss
                #if self.verbose:
                elif k == 'pi_cation':
                    for res in v:
                        #v=[(<Atom instance> HSG1:A:ARG8:CZ, [<Atom instance> IND: : INI 20:C1])]
                        #res=(<Atom instance> HSG1:A:ARG8:CZ, [<Atom instance> IND: : INI 20:C1])
                        #res[0]=<Atom instance> HSG1:A:ARG8:CZ
                        #res[1]=  [<Atom instance> IND: : INI 20:C1]
                        #res[1][0]=  <Atom instance> IND: : INI 20:C1
                        #
                        ss += "USER %s~~%s\n"%(res[0].full_name(), res[1][0].full_name())
                else:
                    for w in v:
                        if verbose: print "@@ getResultStr w=", w
                        try:
                            ss += 'USER ' + w.full_name()+'\n;'
                        except:
                            if verbose: print "except on ", w
                            ss += "USER " + str(w)
                ss += "\n" 
                if verbose: print  "     end ss= ", ss
        return ss


    def buildHydrogenBonds(self, ligand, comment="USER AD> "):
        h_pairDict = self.hydrogen_bond_builder.build(ligand.allAtoms, self.macro_atoms)
        self.h_pairDict = h_pairDict
        #keys should be from lig, values from macro 
        #sometimes are not...@@check this@@
        h_results = {}
        for k, v in h_pairDict.items():
            h_results[k] = 1
            for at in v:
                h_results[at] = 1
        all_hb_ats = AtomSet(h_results.keys())  #all
        d = self.results
        macro_hb_ats = d['macro_hb_atoms'] = all_hb_ats.get(lambda x: x.top==self.macro)
        self.macro_hb_ats = macro_hb_ats
        # process lig
        lig_hb_ats = d['lig_hb_atoms'] = all_hb_ats.get(lambda x: x in ligand.allAtoms)
        self.lig_hb_ats = lig_hb_ats
        outS = comment + "lig_hb_atoms : %d\n"%(len(lig_hb_ats))
        for p in self.lig_hb_ats:  #intD.results['lig_hb_atoms']:
            for hb in p.hbonds:
                if hasattr(hb, 'used'): continue
                if hb.hAt is not None:
                    outS += comment + "%s,%s~%s\n"%(hb.donAt.full_name(), hb.hAt.name, hb.accAt.full_name())
                else:
                    outS += comment + "%s~%s\n"%(hb.donAt.full_name(), hb.accAt.full_name())
                hb.used = 1
                #hsg1V:B:ARG8:NH2,HH22~clean: : INI 20:N5
                #clean: : INI 20:O4,H3~hsg1V:B:ASP29:OD2
                #clean: : INI 20:O4,H3~hsg1V:B:ASP29:OD2
                #clean: : INI 20:N4,H3~hsg1V:B:GLY27:O
                #clean: : INI 20:O2,H2~hsg1V:B:ASP25:OD1
        #macroHStr = self.macro.allAtoms.get(lambda x: hasattr(x, 'hbonds') and len(x.hbonds)).full_name()
        #ligHStr = ligand.allAtoms.get(lambda x: hasattr(x, 'hbonds') and len(x.hbonds)).full_name()
        #return  macroHStr + '==' + ligHStr
        if self.verbose: 
            print  "buildHB returning:"
            print outS
        return outS



    def buildCloseContactAtoms(self, percentCutoff, ligand, comment="USER AD> "):
        pairDict = self.distanceSelector.select(ligand.allAtoms,
                        self.macro_atoms, percentCutoff=percentCutoff)
        self.pairDict = pairDict
        #reset here
        lig_close_ats = AtomSet()
        macro_close_ats = AtomSet()
        cdict = {}
        for k,v in pairDict.items():
            if len(v):
                cdict[k] = 1
            for at in v:
                if at not in macro_close_ats:
                    cdict[at] = 1
        closeAtoms = AtomSet(cdict.keys())
        lig_close_ats = closeAtoms.get(lambda x: x.top==ligand).uniq()
        #ligClAtStr = lig_close_ats.full_name()
        ligClAtStr = comment + "lig_close_ats: %d\n" %( len(lig_close_ats))
        if len(lig_close_ats):
            ligClAtStr += comment + "%s\n" %( lig_close_ats.full_name())
        macro_close_ats = closeAtoms.get(lambda x: x in self.macro_atoms).uniq()
        macroClAtStr = comment + "macro_close_ats: %d\n" %( len(macro_close_ats))
        if len(macro_close_ats):
            macroClAtStr += comment + "%s\n" %( macro_close_ats.full_name())
        #macroClAtStr = "macro_close_ats: " + len(macro_close_ats)+"\n" +macro_close_ats.full_name()
        rdict = self.results
        rdict['lig_close_atoms'] = lig_close_ats
        rdict['macro_close_atoms'] = macro_close_ats
        if self.verbose: print "macroClAtStr=", macroClAtStr
        if self.verbose: print "ligClAtStr=", ligClAtStr
        if self.verbose: print "returning "+ macroClAtStr + '==' + ligClAtStr
        return  macroClAtStr , ligClAtStr


    def getCations(self, atoms):
        #select atoms in ARG and LYS residues
        arg_cations = atoms.get(lambda x: (x.parent.type=='ARG' and \
                                x.name in ['CZ']))
        lys_cations = atoms.get(lambda x: (x.parent.type=='LYS' and \
                                x.name in ['NZ', 'HZ1', 'HZ2', 'HZ3']))
        #select any positively-charged metal ions... cannot include CA here
        metal_cations = atoms.get(lambda x: x.name in ['Mn','MN', 'Mg',\
                                'MG', 'FE', 'Fe', 'Zn', 'ZN'])
        ca_cations = atoms.get(lambda x: x.name in ['CA', 'Ca'] and x.parent.type=='CA')
        cations = AtomSet() 
        #cations.extend(arg_cations)
        for a in arg_cations:
            cations.append(a)
        #cations.extend(lys_cations)
        for a in lys_cations:
            cations.append(a)
        #cations.extend(metal_cations)
        for a in metal_cations:
            cations.append(a)
        #cations.extend(ca_cations)
        for a in ca_cations:
            cations.append(a)
        return cations


    def detectPiInteractions(self, tolerance=0.95, debug=False, use_all_cycles=False):
        if debug: print "in detectPiInteractions"
        self.results['pi_pi'] = []        #stacked rings...?
        self.results['t_shaped'] = []     #one ring perpendicular to the other
        self.results['cation_pi'] = []    #
        self.results['pi_cation'] = []    #
        self.results['macro_cations'] = []#
        self.results['lig_cations'] = []  #
        #at this point have self.results
        if not len(self.results['lig_close_atoms']):
            return
        lig_atoms = self.results['lig_close_atoms'].parent.uniq().atoms
        macro_res = self.results['macro_close_atoms'].parent.uniq()
        if not len(macro_res):
            return
        macro_atoms = macro_res.atoms
        l_rf = RingFinder()
        #Ligand
        l_rf.findRings2(lig_atoms, lig_atoms.bonds[0])
        #rf.rings is list of dictionaries, one per ring, with keys 'bonds' and 'atoms'
        if debug: print "LIG: len(l_rf.rings)=", len(l_rf.rings)
        if not len(l_rf.rings):
            if debug: print "no lig rings found by l_rf!"
            return
        acbs = self.aromatic_cycle_bond_selector
        #acbs = AromaticCycleBondSelector()
        lig_rings = []
        for r in l_rf.rings:
            ring_bnds = r['bonds']
            if use_all_cycles: 
                lig_rings.append(ring_bnds)
            else:
                arom_bnds = acbs.select(ring_bnds)
                if len(arom_bnds)>4:
                    lig_rings.append(arom_bnds)
        if debug: print "LIG: len(lig_arom_rings)=", len(lig_rings)
        self.results['lig_rings'] = lig_rings
        self.results['lig_ring_atoms'] = AtomSet()
        #only check for pi-cation if lig_rings exist
        if len(lig_rings):
            macro_cations = self.results['macro_cations'] = self.getCations(macro_atoms)
            macro_cations = macro_cations.get(lambda x: x.element!='H')
            lig_ring_atoms = AtomSet()
            u = {}
            for r in lig_rings:
                for a in BondSet(r).getAtoms():
                    u[a] = 1
            if len(u): 
                lig_ring_atoms = AtomSet(u.keys())
                lig_ring_atoms.sort()
                self.results['lig_ring_atoms'] = lig_ring_atoms
            if len(macro_cations):
                if debug: print "check distances from lig_rings to macro_cations here"
                #macro cations->lig rings
                pairDict2 = self.distanceSelector.select(lig_ring_atoms,macro_cations)
                z = {}
                for key,v in pairDict2.items():
                    val = v.tolist()[0]
                    if val in macro_cations:
                        z[val] = [key]
                if len(z):
                    self.results['pi_cation'] = (z.items())
                else:
                    self.results['pi_cation'] = []
        #check the distance between the rings and the macro_cations
        self.results['lig_cations'] = self.getCations(lig_atoms)
        lig_cations = self.results['lig_cations'] 
        #remove hydrogens
        lig_cations = lig_cations.get(lambda x: x.element!='H')
        #Macromolecule
        m_rf = RingFinder()
        m_rf.findRings2(macro_res.atoms, macro_res.atoms.bonds[0])
        #rf.rings is list of dictionaries, one per ring, with keys 'bonds' and 'atoms'
        if debug: print "MACRO: len(m_rf.rings)=", len(m_rf.rings)
        if not len(m_rf.rings):
            if debug: print "no macro rings found by m_rf!"
            return
        macro_rings = []
        for r in m_rf.rings:
            ring_bnds = r['bonds']
            if use_all_cycles: 
                macro_rings.append(ring_bnds)
            else:
                arom_bnds = acbs.select(ring_bnds)
                if len(arom_bnds)>4:
                    macro_rings.append(arom_bnds)
        if debug: print "len(macro_arom_rings)=", len(macro_rings)
        self.results['macro_rings'] = macro_rings
        self.results['macro_ring_atoms'] = AtomSet()
        #only check for pi-cation if macro_rings exist
        if len(macro_rings):
            macro_ring_atoms = AtomSet()
            u = {}
            for r in macro_rings:
                for a in BondSet(r).getAtoms(): #new method of bondSets
                    u[a] = 1
            if len(u):
                macro_ring_atoms = AtomSet(u.keys())
                macro_ring_atoms.sort()
                self.results['macro_ring_atoms'] = macro_ring_atoms
            if len(lig_cations):
                if debug: print "check distances from macro_rings to lig_cations here"
                pairDict3 = self.distanceSelector.select(macro_ring_atoms,lig_cations)
                z = {}
                for x in pairDict3.items():
                    #lig cations->macro rings
                    z.setdefault(x[1].tolist()[0], []).append(x[0])
                if len(z):
                    self.results['cation_pi'] = (z.items())
                else:
                    self.results['cation_pi'] = []
                #macro_pi_atoms = AtomSet(pairDict3.keys())
                #l_cations = AtomSet()
                #for v in pairDict3.values():
                #    for x in v:
                #        l_cations.append(x)
                #self.results['cation_pi'] = pairDict3.items()
                #self.results['cation_pi'] = (l_cations, macro_pi_atoms)
        #check for intermol distance <6 Angstrom (J.ComputChem 29:275-279, 2009)
        #compare each lig_ring vs each macro_ring
        for lig_ring_bnds in lig_rings:
            lig_atoms = acbs.getAtoms(lig_ring_bnds)
            lig_atoms.sort()
            if debug: print "len(lig_atoms)=", len(lig_atoms)
            #---------------------------------
            # compute the normal to lig ring
            #---------------------------------
            a1 = Numeric.array(lig_atoms[0].coords)
            a2 = Numeric.array(lig_atoms[2].coords)
            a3 = Numeric.array(lig_atoms[4].coords)
            if debug: print "a1,a2, a3=", a1.tolist(), a2.tolist(), a3.tolist()
            for macro_ring_bnds in macro_rings:
                macro_atoms = acbs.getAtoms(macro_ring_bnds)
                macro_atoms.sort()
                if debug: print "len(macro_atoms)=", len(macro_atoms)
                pD_dist = self.distanceSelectorWithCutoff.select(macro_ring_atoms, lig_atoms, cutoff=self.dist_cutoff)
                if not len(pD_dist[0]):
                    if debug: 
                        print "skipping ligand ring ", lig_rings.index(lig_ring_bnds), " vs ",
                        print "macro ring", macro_rings.index(macro_ring_bnds)
                    continue
                #---------------------------------
                # compute the normal to macro ring
                #---------------------------------
                b1 = Numeric.array(macro_atoms[0].coords)
                b2 = Numeric.array(macro_atoms[2].coords)
                b3 = Numeric.array(macro_atoms[4].coords)
                if debug: print "b1,b2, b3=", b1.tolist(), b2.tolist(), b3.tolist()
                # check for stacking 
                a2_1 = a2-a1
                a3_1 = a3-a1
                b2_1 = b2-b1
                b3_1 = b3-b1
                if debug: print "a2_1 = ", a2-a1
                if debug: print "a3_1 = ", a3-a1
                if debug: print "b2_1 = ", b2-b1
                if debug: print "b3_1 = ", b3-b1
                n1 = crossProduct(a3_1,a2_1) #to get the normal for the first ring
                n2 = crossProduct(b3_1,b2_1) #to get the normal for the second ring
                if debug: print "n1=", n1
                if debug: print "n2=", n2
                n1 = Numeric.array(n1)
                n2 = Numeric.array(n2)
                n1_dot_n2 = Numeric.dot(n1,n2)
                if debug: print "n1_dot_n2", Numeric.dot(n1,n2)
                if abs(n1_dot_n2) >= 1*tolerance: 
                    if debug: print "The rings are stacked vertically" 
                    new_result = (acbs.getAtoms(lig_ring_bnds), acbs.getAtoms(macro_ring_bnds))
                    self.results['pi_pi'].append(new_result)
                if abs(n1_dot_n2) <= 0.01*tolerance: 
                    if debug: print "The rings are stacked perpendicularly" 
                    new_result = (acbs.getAtoms(lig_ring_bnds), acbs.getAtoms(macro_ring_bnds))
                    self.results['t_shaped'].append(new_result)


    def get_pi_pi(self, print_ctr=1, comment="USER AD> "):
        #one ring parallel to another 
        #@@ UNTESTED! ... need test data
        if not len(self.results.get( 'pi_pi')):
            return ""
#        self.results['pi_pi'] = []        #stacked rings...?
        res = self.results['pi_pi']
        ss = comment + "pi_pi: %d\n"%(len(self.results['pi_pi']))
        ####
        #3l1m_lig_vs:A:HEM150:C4A,C3D,C1A,C2D,C4D,C2A,NA,ND,C3A,C1D : 3l1m_rec:A:PHE65:CD2,CD1,CG,CZ,CE1,CE2
        #3l1m_lig_vs:A:HEM150:C3B,NB,C4A,C1A,C4B,NA,C2A,C2B,C3A,C1B : 3l1m_rec:A:PHE65:CD2,CD1,CG,CZ,CE1,CE2
        #3l1m_lig_vs:A:HEM150:C3B,C1B,C3C,NB,C4C,NC,C2B,C1C,C4B,C2C : 3l1m_rec:A:PHE65:CD2,CD1,CG,CZ,CE1,CE2
        #3l1m_lig_vs:A:HEM150:NC,C4C,C1C,C3C,C2C : 3l1m_rec:A:PHE65:CD2,CD1,CG,CZ,CE1,CE2
        #3l1m_lig_vs:A:HEM150:C1A,NA,C2A,C3A,C4A : 3l1m_rec:A:PHE65:CD2,CD1,CG,CZ,CE1,CE2
        #3l1m_lig_vs:A:HEM150:NB,C2B,C3B,C1B,C4B : 3l1m_rec:A:PHE65:CD2,CD1,CG,CZ,CE1,CE2
        #3l1m_lig_vs:A:HEM150:NC,C4C,C1C,C3C,C2C : 3l1m_rec:A:PHE65:CD2,CD1,CG,CZ,CE1,CE2
        #3l1m_lig_vs:A:HEM150:C4D,ND,C1D,C3D,C2D : 3l1m_rec:A:PHE65:CD2,CD1,CG,CZ,CE1,CE2
        #
        # build string for  self.results['pi_pi'] 
        for res in self.results['pi_pi']:
            ss += comment + "%s~~%s\n"%(res[0].full_name(), res[1].full_name())
            ###ss += "USER %s~~%s\n"%(res[0].full_name(), res[1].full_name())
            #ss += "USER %s~~%s\n"%(res[0].full_name(), res[1][0].full_name())
        #print "returning ss=" , ss
        return ss


    def get_t_shaped(self, print_ctr=1, comment="USER AD> "):
        #one ring perpendicular to the other
        #@@ UNTESTED! ... need test data
        if not len(self.results.get( 't_shaped')):
            return ""
        #print "in gpcr: self.results[t_shaped]=", self.results['t_shaped']
        ss = comment + "t_shaped: %d\n"%(len(self.results['t_shaped']))
        # build string for  self.results['t_shaped'] 
        for res in self.results['t_shaped']:
            #for example:
            #????
            #????
            #????
            #
            ss += comment + "%s~~%s\n"%(res[0].full_name(), res[1][0].full_name())
        return ss


    def get_cation_pi(self, print_ctr=1, comment="USER AD> "):
        #cation near aromatic ring 
        #@@ UNTESTED! ... need test data
        if not len(self.results.get( 'cation_pi')):
            return ""
        ss = comment + "cation_pi: %d\n"%(len(self.results['cation_pi']))
        # build string for  self.results['cation_pi'] 
        for res in self.results['cation_pi']:
            #for example:
            #????
            #????
            #????
            #
            ss += comment + "%s~~%s\n"%(res[0].full_name(), res[1][0].full_name())
        return ss


    def get_pi_cation_result(self, print_ctr=1, comment="USER AD> "):
        #aromatic ring near cation 
        #print "in gpcr: self.results[pi_cation]=", self.results['pi_cation']
        if not len(self.results.get( 'pi_cation')):
            return ""
        ss = comment + "pi_cation: %d\n"%(len(self.results['pi_cation']))
        # build string for self.results['pi_cation'] 
        for res in self.results['pi_cation']:
            #for example:
            #v=[(<Atom instance> HSG1:A:ARG8:CZ, [<Atom instance> IND: : INI 20:C1])]
            #res=(<Atom instance> HSG1:A:ARG8:CZ, [<Atom instance> IND: : INI 20:C1])
            #res[0]=<Atom instance> HSG1:A:ARG8:CZ
            #res[1]=  [<Atom instance> IND: : INI 20:C1]
            #res[1][0]=  <Atom instance> IND: : INI 20:C1
            #
            #attempt to get names for all atoms in the whole ring:
            res1_str = res[1][0].full_name()
            for rr in self.results['lig_rings']:
                ats = rr.getAtoms()
                if res[1][0] in ats:
                    res1_str = ats.full_name()
                    break
            ss += comment + "%s~~%s\n"%(res[0].full_name(), res1_str)
            #ss += "USER %s~~%s\n"%(res[0].full_name(), res[1][0].full_name())
        return ss



    def print_ligand_residue_contacts(self, print_ctr=1):
        ctr = 1
        #pairDict is atom-based 
        #for residue-based report:
        # need to build lists of unique parents of keys 
        # and lists of unique parents of corresponding values
        res_d = {}
        for at in self.pairDict.keys():
            if at.parent not in res_d.keys():
                res_d[at.parent] = {}
            for close_at in self.pairDict[at]:
                res_d[at.parent][close_at.parent] = 1
        #print it out
        for lig_res in res_d.keys():
            if print_ctr:
                print ctr, lig_res.parent.name+':'+ lig_res.name + '->',
            else:
                print lig_res.parent.name+':'+ lig_res.name + '->',
            for macro_res in res_d[lig_res]:
                print macro_res.parent.name + ':' + macro_res.name + ',',
            print
            ctr += 1
        return res_d


    def print_macro_residue_contacts(self, print_ctr=1):
        ctr = 1
        #pairDict is atom-based 
        #for residue-based report:
        # need to build lists of unique parents of keys 
        # and lists of unique parents of corresponding values
        res_d = {}
        for at_key, at_list in self.pairDict.items():
            for at in at_list:
                if at.parent not in res_d.keys():
                    res_d[at.parent] = {}
                res_d[at.parent][at_key.parent] = 1
        #print it out
        for macro_res in res_d.keys():
            if print_ctr:
                print ctr, macro_res.parent.name+':'+ macro_res.name + '->',
            else:
                print macro_res.parent.name+':'+ macro_res.name + '->',
            for lig_res in res_d[macro_res]:
                print lig_res.parent.name + ':' + lig_res.name + ',',
            print
            ctr += 1
        return res_d

    def print_report(self, keylist=[]):
        if not len(keylist):
            keylist = [
                'lig_close_atoms',
                'lig_hb_atoms',
                'lig_hbas',
                'macro_close_atoms',
                'macro_hb_atoms',
                    ]
        d = self.results
        if self.verbose:
            for k in keylist:
                print k, ':', len(d[k]), '-', d[k].__class__


    def print_hb_residue(self, print_ctr=1):
        ctr = 1
        #pairDict is atom-based 
        res_d = {}
        for at in self.h_pairDict.keys():
            if at.parent not in res_d.keys():
                res_d[at.parent] = {}
            for close_at in self.h_pairDict[at]:
                res_d[at.parent][close_at.parent] = 1
        # print it out
        # Hbond instance are define with donAt,hAt ~ accAtt  (tilde)
        for don_res in res_d.keys():
            if print_ctr:
                print ctr, don_res.top.name+':'+don_res.parent.name+':'+ don_res.name + '->',
            else:
                print don_res.top.name+':'+don_res.parent.name+':'+ don_res.name + '->',
            for acc_res in res_d[don_res]:
                print acc_res.top.name+':'+acc_res.parent.name + ':' + acc_res.name + ',',
            print
            ctr += 1
        return res_d




## Automatically adapted for numpy.oldnumeric Jul 23, 2007 by 

#############################################################################
#
# Author: Ruth HUEY, Michel F. SANNER
#
# Copyright: M. Sanner TSRI 2002
#
#############################################################################


# $Header: /opt/cvs/python/packages/share1.5/AutoDockTools/DlgParser.py,v 1.71 2011/12/19 18:02:36 rhuey Exp $
#
# $Id: DlgParser.py,v 1.71 2011/12/19 18:02:36 rhuey Exp $
#
#
#
#
#
#
#

"""
This Object parses the result of an AutoDock job and returns a dictionary. 

"""
import os
from string import find, join, replace, split, rfind, strip
import re
import numpy.oldnumeric as Numeric

from AutoDockTools.ResultParser import ResultParser


class DlgParser(ResultParser):
    """ reads log from a AutoDock docking and return structured data"""

    keywords = ResultParser.keywords + [
        'coords',
        'vdw_energies',
        'estat_energies',
        'total_energies',   #vdw_energies+estat_energies
        'inhib_constant',
        'intermol_energy',  #(1) 
        'internal_energy',  #(2) NB: 1+2->final docked energy
        'torsional_energy', #(3) NB: 1+3->free energy of binding
        'run',
        'parameter_file',   #AD4 parameter library filename
        'include_1_4_interactions',   #AD4 internal energy switch
        ]
        

    def __init__(self, dlgFile=None):
        """selected dlgFile,ok sets which docked conformations to show"""
        ResultParser.__init__(self)
        self.filename = dlgFile
        self.WARNINGS = []
        self.population_list = []
        self.population = []
        #set up dict here
        self.getReDict()
        self.ligand_atom_count = 0
        self.ligand_count = 0
        if dlgFile:
            self.filename = os.path.basename(dlgFile)
            self.parse(dlgFile)



    def parse(self, filename):
        """
        uses keys:
            '|           AutoDock' 
            'DPF> outlev'
            'DPF> '
            'INPUT-PDBQ: '
            'DOCKED: '
            'State='
            '^Seeds'
            '^   [1-9]
        after parsing: 
            self.outlev is switch to correctly find coord fields
            self.clist is list of dictionaries for states,
            self.dpfLines are for building DockingParameters object
            self.histogramlines show results histogram
            self.clusterlines have clustering info
            self.modelList is list of dicts of docked: coords plus energy info 
            self.wroteAll is a flag set by 'write_all_cluster_members' or
                    by 'write_all'
        """
        self.filename = filename
        self.version = 4.2
        #reset
        for item in ['clist','clusterlines','dpfLines','histogramlines',\
                     'ligLines','modelList', 'initial_clist']:
            setattr(self, item, [])

        self.wroteAll = 0
        self.clusterRecord = None
        self.getReDict()
        try:
            dlgptr = open(filename, 'r')
        except:
            raise IOError
        allLines = dlgptr.readlines()
        self.allLines = allLines
        self.match(allLines)
        if self.version>=4.0:
            ad4_keywds = ['vdw_energy', 'estat_energy', 
                          'inhib_constant_units',
                          'vdw_hb_desolv_energy',
                          'electrostatic_energy',
                          'moving_ligand_fixed_receptor',
                          'moving_ligand_moving_receptor',
                          'total_internal',
                          'ligand_internal',
                          'receptor_internal',
                          'unbound_energy',
                          ]
            self.keywords.extend(ad4_keywds)


    def getReDict(self):
        if hasattr(self, 'reDict'):
            for k, d in self.reDict.items():
                d['lines'] = []
            return
        self.reDict = {}
        #had to add MODEL|^USER|^ATOM key for outlev -1
        #seems to slow stuff down alot
        self.compute_unbound_extended = False
        self.reKeys = [
            '                 |            AutoDock', #used to correct autodock4 output 
            'DPF> outlev',       #format switch
            'DPF> compute_unbound_extended', # affects number of seeds
            'DPF> write_all',    #wroteAll switch
            'INPUT-PDBQ: USER    NEWDPF',         #extra info in case of clustering dlg
            'DPF> ',             #lines for dpo
            'INPUT-PDBQ: ',      #lines for ligand
            'INPUT-PDBQT: ',     #lines for ligand, v4.0_old_version
            'INPUT-LIGAND-PDBQT: ',     #lines for ligand, v4.0
            'INPUT-FLEXRES-PDBQT: ',     #lines for flex_res, v4.0
            'DOCKED: ',          #lines for models
            '^MODEL|^USER|^ATOM|^ENDMDL',       #outlev -1 modellines
            'State=',            #lines for conformations
            '^ [1-9]|^  [1-9]|^   [1-9]',   #cluster and histogram lines
            '^Seeds',            #lines for seed
            'Total number of atoms found in PDBQT file', #to get ligand at #
            '^Atom: ID',         #for non-bond table
            'WARNING',           #accumulate all warning messages
            'Energy= ',          #accumulate initial states available if outlev>1
            '<population size=',
            ".*autodock.*: Successful Completion",
            "Coordinates of Central Grid Point of Maps",
            ]
        self.reFuncs = [ 
            self.setADVersion,
            self.setOutlev,
            self.setComputeUnboundExtended,
            self.setWroteAll,
            self.getNewDpfInfo,
            self.processDpfLines,
            self.processLigLines,
            self.processLigLinesV4,
            self.processLigLinesV4,
            self.processFlexResLinesV4,
            self.getModelLines,
            self.getShortModelLines,
            self.getDlgStates,
            self.getClusterInfo,
            self.getSeedInfo,
            self.getLigandAtomCount,
            self.getNonBondTable,
            self.getWARNINGS,
            self.getInitialConfs,
            self.getPopulations,
            self.getTime,
            self.getCenterMapPt,
            ]

        for i in range(len(self.reKeys)):
            k = self.reKeys[i]
            dict =  self.reDict[k] = {}
            dict['re'] = re.compile(k)
            dict['lines'] = []
            dict['func'] = self.reFuncs[i]


    def match(self, allLines, verbose=False):
        self.getReDict()
        self.tested = 1
        for i in range(len(allLines)):
            item = allLines[i]
            #if find item, mark it found + don't test
            for k in self.reKeys:
                d = self.reDict[k]
                m = d['re'].match(item)
                if m:
                    d['lines'].append(item) 
                    break
            #process Energy= separately
            #energy_key = 'Energy= '
            energy_key = "Energy= "
            d = self.reDict[energy_key]
            if item.find(energy_key)>-1:
                d['lines'].append(item)
                d['lines'].append(allLines[i+1])
        for k in self.reKeys:
            d = self.reDict[k]
            lines = d['lines']
            if lines==[]:
                input_key = 'INPUT-PDBQ: '
                if k==input_key:
                    if verbose and self.version<4.0:
                        print "!!no lines found for key=", k, "!!"
                else:
                    if verbose:
                        print "!!no lines found for key=", k, "!!"
            else:
                apply(d['func'], (lines,), {})


    def setADVersion(self,lines):
    	#May need to modify for next autodock version -Huameng
        if len(lines):
            for l in lines:
                if find(l, 'AutoDock')>-1:
                    ll = split(l)
                    version = ll[2]
                    if len(version)>3:
                        version = version[:3]
                    self.version = float(version)
                    break
        else:
            self.version = 3.0


    def setOutlev(self,lines):
        if len(lines):
            ll = split(lines[0])
            self.outlev = int(ll[2])
        else:
            self.outlev = None


    def setComputeUnboundExtended(self,lines):
        if len(lines):
            self.compute_unbound_extended = True


    def setWroteAll(self, lines):
        if len(lines):
            self.wroteAll = 1


    def getLigandAtomCount(self, lines):
        self.ligand_atom_count = int(split(lines[0])[-2])
        #print "parsed ligand_atom_count =", self.ligand_atom_count


    def getWARNINGS(self, lines):
        #print "found ", len(lines), ' WARNINGS'
        self.WARNINGS = lines


    def getInitialConfs(self, lines):
        #print "found ", len(lines), ' InitialConfs'
        self.InitialConfs= lines
        x = len(lines)/2
        for j in range(x):
            energy = float(lines[2*j].strip().split()[-1])
            clist = map(float, lines[2*j+1].strip().split())
            self.initial_clist.append([energy, clist])


    def getCenterMapPt(self, lines):
        for l in lines:
            if l.find('Coordinates of Central Grid Point of Maps')>-1:
                ll = split(l, '=')
                str_pt = ll[1].strip()  #"(13.231, 56.852, 54.080)
                #str_pt = ll[1]  #"(13.231, 56.852, 54.080)
                self.center_pt = map(float, str_pt[2:-2].split(','))


    def getTime(self, lines, echo=False):
        ind = self.allLines.index(lines[-1])
        #print "ind=", ind
        self.info_line = None
        self.total_time = 0
        for info_line in self.allLines[ind+1:]:
            #print "testing info_line->", info_line
            #print "info_line.find('Real=')=", info_line.find('Real=')
            if info_line.find("Real=")>-1:
                #print "FOUND IT!!"
                self.info_line = info_line
                break
        if self.info_line:
            #Real= 23m 38.38s,  CPU= 22m 08.78s,  System= 0.49s
            time_string = self.info_line.split(',')[0][6:]
            #23m 38.38s
            time_list = time_string.split()
            #['5h', '14m', '39.77s']
            #['14m', '39.77s']
            #['39.77s']
            scale_factor ={ 'h':60*60,'m':60,'s':1}
            for x in time_list:
                #['5h', '14m', '39.77s']
                self.total_time += scale_factor[x[-1]] * float(x[:-1])
        

    def getPopulations(self, lines, echo=False):
        currently_build_pop = True
        ind = self.allLines.index(lines[0])
        if echo: print "found start of population at index=", ind
        found = False
        all_population_lines = []
        plist = []
        currently_building_pop = True
        for line in self.allLines[ind:]:
            if line.find("</pop")==0:
                if echo: print 'found end of a population', ind
                found = True
                all_population_lines.append(plist)
                if echo: print "added plist:", len(plist)
                currently_building_pop = False
                plist = []
                #break
            elif line.find("<pop")==0:
                if echo: print 'found start of a population', ind
                currently_building_pop = True
                plist = []
            elif currently_building_pop:
                plist.append(line)
        x = len(plist)
        if echo: print x, 'lines in getInitialPopulation'
        self.population_list = []
        #figure out number of torsions:
        num_values = len(all_population_lines[0][0].strip().split())
        # 0       1       2   3      4      5      6      7     8     9 # ....
        #index   enrgy   age  trnX  trnY   trnZ    qtX    qtY   qtZ  qtANG [tor1 tor2 ...]
        #150	3.36e+05  0	 40.996 13.492 23.851  0.092 -0.658 0.748 9.890  
        num_torsions = num_values - 10
        for plist in all_population_lines:
            #previously
            #150	3.36e+05	40.996 13.492 23.851  0.092 -0.658 0.748 9.890  
            #current ad4 version has age between energy and tran_x, eg 0, third number
            #150	3.36e+05  0	 40.996 13.492 23.851  0.092 -0.658 0.748 9.890  
            current_pop = []
            for line in plist:
                ll = line.strip().split()
                ind_index = int(ll[0])
                age  = int(ll[2])
                clist = map(float, ll[3:])
                cdict = {}
                cdict['binding_energy'] = float(ll[1])
                cdict['age'] = int(ll[2])
                cdict['trn_x'] = float(ll[3])
                cdict['trn_y'] = float(ll[4])
                cdict['trn_z'] = float(ll[5])
                cdict['qtn_nx'] = float(ll[6])
                cdict['qtn_ny'] = float(ll[7])
                cdict['qtn_nz'] = float(ll[8])
                cdict['qtn_ang_deg'] = float(ll[9])
                cdict['num_torsions'] = num_torsions
                if num_torsions>0:
                    cdict['torsion_values'] = clist[7:]
                current_pop.append(cdict)
            self.population_list.append(current_pop)
        self.initial_pop = self.population_list[0]
        self.population = self.initial_pop
            

    def getNonBondTable(self, lines, echo=False):
        ind = self.allLines.index(lines[0])
        ct = self.ligand_atom_count
        nb_lines = self.allLines[ind+2:ind+2+ct]
        self.nb_array = Numeric.zeros((ct, ct))
        if echo:
            for l in nb_lines:
                print l,
        for i in range(ct):
            l = nb_lines[i][10:-1]
            for j in range(ct):
                jind = j*2 + 1
                if l[jind]=='X':
                    self.nb_array[i][j]=1
            if echo:
                print
                print l
                print self.nb_array[i]


    def getSeedInfo(self, lines):
        seeds = []
        for l in lines:
            ilist = split(l)
            slist = []
            for item in ilist[1:]:
                slist.append(int(item))
            seeds.append(tuple(slist))
            #seeds.append(tuple((int(ilist[1]),int(ilist[2]))))
        if self.version>=4.0 and self.compute_unbound_extended:
            #there are seeds for compute_unbound_extended
            #remove the first entry
            seeds = seeds[1:]
        if len(seeds)==len(self.clist):
            for i in range(len(seeds)):
                conf = self.clist[i]
                if conf.has_key('run') or self.version>=4.0:
                    #outlev -1 prints States in order
                    s = seeds[i]
                    conf['rseed1'] = s[0]
                    if len(s)==2:
                        conf['rseed2'] = s[1]
                    #elif not conf.has_key('run'):
                #else:
                #    continue
                    #outlev -1 prints States in order
                    #s = seeds[i]
                    #conf['rseed1'] = s[0]
                    #if len(s)==2:
                        #conf['rseed2'] = s[1]
                #else:
                    #run = conf['run']
                    ##run is 1-based; seeds indexed starting w/0
                    #s = seeds[run-1]
                    #conf['rseed1'] = s[0]
                    #if len(s)==2:
                        #conf['rseed2'] = s[1]
                #else:
                #run = conf['run']
                ##run is 1-based; seeds indexed starting w/0
                #s = seeds[run-1]
                #conf['rseed1'] = s[0]
                #if len(s)==2:
                #    conf['rseed2'] = s[1]


    def getClusterInfo(self, lines):
        cl = self.clusterlines
        hl = self.histogramlines
        for l in lines:
            if find(l, 'RANKING')>-1:
                cl.append(l[:-1])
            elif find(l, '#')>-1:
                hl.append(l[:-1])
        if len(cl):
            self.getClusterRecord(cl)
        else:
            self.clusterRecord = None


    def getClusterRecord(self, cl):
        #print 'in getClusterRecord'
        clRecList = []
        curList = []
        #curList gets list of conf info
        curInd = int(split(cl[0])[0])
        ctr = 1
        for l in cl:
            ll = split(l)
            #when built, newList is
            #[Rank,SubRank,Run,DockedEnergy,ClusterRMSD,RefREMSD]
            newList = map(lambda x:int(x),ll[:3])
            #3/29/05
            if self.wroteAll and self.version<4.0:
                #print "setting run number to ", ctr
                newList[2] = ctr
                ctr = ctr + 1
            newList2 = map(lambda x:float(x),ll[3:-1])
            newList.extend(newList2)
            if newList[0]==curInd:
                curList.append(newList)
            else:
                clRecList.append(curList)
                curList = [newList]
                curInd = newList[0]
        clRecList.append(curList)
        self.clusterRecord = clRecList
        
            

    def processDpfLines(self, lines):
        self.recluster = 0
        self.mlsd_count = 0
        for l in lines:
            if len(l[:-1])>5:
                #check for any garbage in dpf
                if l.find('DPF> #')==0:
                    continue 
                if l.find('DPF> cluster')>-1:
                    self.recluster = 1
                if l.find('move')>-1:
                    self.ligand_count +=1 #for wcg results 
                    #print " now self.ligand_count=", self.ligand_count
                    #check for run_on ligand_filename here
                    if l.find('pdbqt#')>-1:
                        ind = l.find('pdbqt#')
                        l = l[:ind+5] + ' ' + l[ind+5:]
                if l.find('ligand')>-1: 
                    self.mlsd_count +=1 #mlsd_docking
                self.dpfLines.append(l[5:-1])
                if l.find('ga_run')>-1:
                    ll = split(l)
                    self.runs = int(ll[2])
                    #print "self.runs=", self.runs
                elif l.find('runs')>-1 and l.find('simanneal')<0:
                    ll = split(l)
                    self.runs = int(ll[2])
                    #print "SA: self.runs=", self.runs

    
    def getNewDpfInfo(self, lines):
        if not len(lines):
            return
        dpfLines = []
        keys = []
        for l in lines:
            ll = split(l)
            ind = ll.index('NEWDPF')
            k = ll[ind+1]
            #only add each key once
            if k not in keys:
                dpfLines.append(join(ll[ind+1:]))
                keys.append(k)
        if len(dpfLines):
            self.dpfLines.extend(dpfLines)


    def processLigLines(self, lines):
        #do not use this for version 4.0
        if self.version>=4.0:
            return
        ligLINES = []
        foundRun = 0
        for l in lines:
            #in clustering dlg, multiple copies of input-pdbq are present
            if find(l, ' Run')>-1 and foundRun:
                break
            elif find(l, ' Run')>-1:
                foundRun = 1
            else:
                ligLINES.append(l[12:-1])
        #check here to remove lines of just spaces
        nl = []
        for l in ligLINES:
            if len(strip(l)):
                nl.append(l)
        self.ligLines = nl
        #print "in processLigLines:len(ligLines)=", len(nl)
        #self.ligLines = ligLINES


    def processFlexResLinesV4(self, lines):
        #print "in processFlexResLinesV4: len(self.ligLines=)", len(self.ligLines)
        if self.version<4.0:
            print "not version 4.0! RETURNING!!"
            return
        ligLINES = []
        foundRun = 0
        ind = 21
        for l in lines:
            #in clustering dlg, multiple copies of input-pdbq are present
            if find(l, 'Run')>-1 and foundRun:
                break
            elif find(l, 'Run')>-1:
                foundRun = 1
            elif find(l, '^_____________________')>-1:
                #last line is ________________-
                break
            else:
                ligLINES.append(l[ind:-1])
        #check here to remove lines of just spaces
        nl = []
        for l in ligLINES:
            if len(strip(l)):
                nl.append(l)
        self.flex_res_lines = nl
        #print "end pFRLV4: len(self.flex_res_lines)=", len(nl)
        #print "end processFlexResLinesV4: len(self.ligLines=)", len(self.ligLines)
        self.hasFlexRes = True
        self.flex_res_count = nl.count("REMARK  status: ('A' for Active; 'I' for Inactive)")

    
    def processLigLinesV4(self, lines, verbose=False):
        #use this only for versions 4.0 and greater
        #hack to fix run-on INPUT-PDBQT: line from wcg 
        if verbose: print " self.ligand_count =", self.ligand_count
        first_line = lines[0]
        if len(first_line)>100 and first_line.count("INPUT-PDBQT:")>2:
            lines = first_line.replace("INPUT-PDBQT:", "\nINPUT-PDBQT:").split('\n')
            #exclude leading '' and trailing '' from split
            lines = lines[1:-1]
            #print "replaced lines with ", len(lines), ' lines'
        if self.version<4.0:
            print "ERROR in processLigLinesV4: self.version=", self.version, "<4.0"
            return
        ligLINES = []
        foundRun = 0
        if find(lines[0], 'INPUT-LIGAND-PDBQT')==0:
            ind = 20
        elif find(lines[0], 'INPUT-PDBQ')==0:
            ind = 13
        for l in lines:
            #in clustering dlg, multiple copies of input-pdbq are present
            if find(l, 'Run')>-1 and foundRun:
                break
            elif find(l, 'Run')>-1:
                foundRun = 1
            elif find(l, 'TORSDOF')>-1:
                #eg run-on TORSDOF line: 
                #TORSDOF 3___
                lastChar = l.find('_')
                #previously l[13:-1]
                #ligLINES.append(l[13:lastChar])
                ligLINES.append(l[ind:lastChar])
                l_index = lines.index(l)
                if l_index==len(lines)-1 or self.ligand_count>1:
                    if verbose: print "found TORSDOF on last line!"
                    break
                next_line = lines[l_index+1]
                if find(next_line, 'BEGIN_RES')<0 and find(next_line, 'active torsions:')<0: #hack for multiple ligands
                    if self.recluster==0:
                        print len(lines[l_index:]), ' lines left unparsed!!!'
                        break
                    else:   
                        print "@@This dlg file is a RE-CLUSTERING file@@"
            else:
                #ligLINES.append(l[13:-1])
                ligLINES.append(l[ind:-1])
         
        lig_ct = ligLINES.count("REMARK  status: ('A' for Active; 'I' for Inactive)")
        if verbose: print '2: INPUT_LIGAND number from ligLINES = ', lig_ct
        #check here to remove lines of just spaces
        nl = []
        for l in ligLINES:
            if len(strip(l)):
                nl.append(l)
        #print "end pLV4: len(self.ligLines)=", len(nl)
        self.ligLines = nl


    
    #######################################
    #####
    ##### getModelLines(self, lines)
    #####
    #######################################    
    def getModelLines(self, lines):
        #print "in getModelLines with ", len(lines), ' lines'
        if not len(lines):
            return
        modelList = []
        if find(lines[0], 'DOCKED')==0:
            ind = 8
        elif find(lines[0], 'INPUT-LIGAND-PDBQT')==0:
            ind = 20
        elif find(lines[0], 'INPUT-PDBQ')==0:
            ind = 12
        #preprocess lines to remove DOCKED or INPUT-PDBQ
        nlines = []
        ctr = 0
        self.run_models = {}
        has_docked = False
        for l in lines:
            #if self.version!=4.0:
            #    nlines.append(l[ind:-1])
            #else:
            #    if find(l, 'ATOM')>-1:
            #        newLine = l[ind:64] + l[66:-1]
            #        nlines.append(newLine)
            #    else:
            #        nlines.append(l[ind:-1])
            if self.version>=4.0:
                if l.find("DOCKED: USER    Run = ")>-1:
                    ll = l.split()
                    ctr = int(ll[4])
                    #print "found run ", ctr
                    has_docked = True
                    nlines = []
                elif l.find("DOCKED: ENDMDL")>-1:
                    self.run_models[ctr] = nlines
                    #print "saved run ", ctr, " nlines=", len(nlines)
                    has_docked = False
                    ctr = -1
                elif has_docked is True:
                    #print 'appending ', l
                    if l.find("ATOM")>-1:
                        try:
                            nlines.append(l[ind:-1])
                        except:
                            print 'cutting out 64+65 of ', l
                            nlines.append(l[ind:64]+ l[66:-1])
                    else:
                        nlines.append(l[ind:-1])
            else:
                nlines.append(l[ind:-1])
        #print "len(nlines)=", len(nlines)
        if self.version<4.0:
            curMod = [nlines[0]]
            for l in nlines[1:]:
                if find(l, 'MODEL')>-1:
                    modelList.append(curMod)
                    curMod = [l]
                else:
                    curMod.append(l)
            modelList.append(curMod)
            #self.makeModels(modelList)
        else:
            #print "setting modelList to ", self.run_models.values()
            modelList = []
            for i in range(self.runs):
                try:
                    modelList.append(self.run_models[i+1])
                except:
                    print "unable to load expected model number ", i+1
        self.modelList = modelList
        self.makeModels(modelList)


    def getShortModelLines(self, lines):
        if not len(lines):
            return
        modelList = []
        curMod = [lines[0][:-1]]
        endmdl = 0
        for l in lines[1:]:
            #use endmdl because cluster dlg format
            # has MODEL then lines w/o MODEL then MODEL
            #again in desc of 1 model...
            if find(l, 'ENDMDL')>-1:
                curMod.append(l)
                modelList.append(curMod)
                endmdl = 1
            elif endmdl:
                curMod = [l[:-1]]
                endmdl = 0
            else:
                curMod.append(l[:-1])
        #don't append after for loop 
        #because it happens in the else
        #print 'len(modelList)=', len(modelList)
        self.modelList = modelList
        self.makeModels(modelList)


    def getDlgStates(self, lines):
        #print "in getDlgStates: len(self.clist)=", len(self.clist)
        if len(self.reDict['^MODEL|^USER|^ATOM|^ENDMDL']['lines']):
            #print 'not building states because models present'
            return
        if len(self.clist)==self.runs:
            for i in range(len(self.clist)):
                cl = self.clist[i]
                if not cl.has_key('run'):
                    cl['run'] = i+1
            return
        else:
            #print "resetting self.clist
            self.clist = []
        for l in lines:
            # in test-1 State= + 17 items: 3 trans, 4quat + ndihe(10) torsions
            xx = split(l)
            # remove possible punctuation
            for ind in range(len(xx)):
                if xx[ind][-1]==',': 
                    xx[ind] = xx[ind][:-1]
                    raise 'comma'
                if xx[ind][-1]=='.': 
                    xx[ind] = xx[ind][:-1]
                    raise 'period'
                
            #transList = xx[1:4]
            trans = []
            for p in [0,1,2]:
                trans.append(float(xx[p+1]))
            trans = tuple(trans)

            quat = []
            for n in xx[4:8]:
                quat.append(float(n))
            quat = tuple(quat)

            angList = []
            #NB: here torsions are in the same line
            for n in xx[8:]:
                angList.append(float(n))

            #BUILD A DICTIONARY and put it in clist
            d = {}
            d['trn_x'] = trans[0]
            d['trn_y'] = trans[1]
            d['trn_z'] = trans[2]
            d['qtn_nx'] = quat[0]
            d['qtn_ny'] = quat[1]
            d['qtn_nz'] = quat[2]
            d['qtn_ang_deg'] = quat[3]
            d['num_torsions'] = len(angList)
            d['torsion_values'] = angList
            self.clist.append(d)
            #print "added ", len(self.clist), "th conformation"


    def makeModels(self, modelList):
        #print "in makeModels with ", len(modelList), ' models to build'
        #print "self.version=", self.version
        if hasattr(self, 'clist') and len(self.clist):
            #print "parser already has clist, returning!"
            return
        #print "##########  SETTING clist to [] ###########"
        clist = []
        ctr = 1
        for curMod in modelList:
            clist.append(self.makeModel(curMod))
            #@@@CHECK THIS@@@
            clist[-1]['run'] = ctr
            ctr +=1
            #print "now len(clist)=", len(clist)
        self.clist = clist
        #print "end of  makeModels: len(self.clist)=", len(self.clist)


    def makeModel(self, lines):
        coords = []
        vdW = []
        Elec = []
        d = {}
        corr = 0
        if hasattr(self, 'outlev') and self.outlev==-1:
            corr = -1
        binding_energy2 = None
        version = self.version
        for l in lines:
            ll = split(l)
            #if find(l, 'MODEL')>-1:
            #    d['num'] = int((ll)[1])
            if find(l, 'Run')>-1 and find(l, 'Rank')==-1:
                d['run'] = int((ll)[3])
            elif find(l, 'Estimated Free Energy of Binding')>-1:
                d['binding_energy'] = float((ll)[7])
                if version>=4.0: d['energy'] = float((ll)[7])
            elif find(l, 'vdW + Hbond + desolv Energy')>-1:
                d['vdw_hb_desolv_energy'] = float((ll)[8])
            elif find(l, 'Electrostatic Energy')>-1:
                if (ll[1] == 'Electrostatic'):
                    d['electrostatic_energy'] = float((ll)[4])
                elif (ll[1] == 'Intermol.'):
                    d['electrostatic_energy'] = float((ll)[5])
            elif find(l, 'Moving Ligand-Fixed Receptor')>-1:
                d['moving_ligand_fixed_receptor'] = float((ll)[5])
            elif find(l, 'Moving Ligand-Moving Receptor')>-1:
                d['moving_ligand_moving_receptor'] = float((ll)[5])
            elif find(l, 'Total Internal Energy')>-1:
                d['total_internal'] = float((ll)[7])
            elif find(l, 'Internal Energy Ligand')>-1:
                d['ligand_internal'] = float((ll)[5])
            elif find(l, 'Internal Energy Receptor')>-1:
                d['receptor_internal'] = float((ll)[5])
            elif find(l, 'Torsional Free Energy')>-1:
                d['torsional_energy'] = float((ll)[6])
            elif find(l, 'Estimated Inhibition Constant')>-1:
                if find(l, 'N/A') < 0:
                    d['inhib_constant'] = float((ll)[6])
                    if version>=4.0:
                        d['inhib_constant_units'] = ll[7]
            elif find(l, 'Final Docked Energy')>-1:
                d['docking_energy'] = float((ll)[5])
            elif find(l, 'Final Intermolecular Energy')>-1:
                d['intermol_energy'] = float((ll)[6])
            elif find(l, 'Final Internal Energy of Ligand')>-1:
                d['internal_energy'] = float((ll)[8])
            elif find(l, 'Final Internal Energy')>-1:
                d['internal_energy'] = float((ll)[6])
            elif find(l, 'Torsional Free Energy')>-1:
                d['torsional_energy'] = float((ll)[6])
            elif find(l, 'NEWDPF tran0')>-1:
                d['trn_x'] = float(ll[3])
                d['trn_y'] = float(ll[4])
                d['trn_z'] = float(ll[5])
            elif find(l, 'NEWDPF quat0')>-1:
                d['qtn_nx'] = float(ll[3])
                d['qtn_ny'] = float(ll[4])
                d['qtn_nz'] = float(ll[5])
                d['qtn_ang_deg'] = float(ll[6])
            elif find(l, 'NEWDPF quaternion0')>-1:
                d['quaternion_nx'] = float(ll[3])
                d['quaternion_ny'] = float(ll[4])
                d['quaternion_nz'] = float(ll[5])
                d['quaternion_nw'] = float(ll[6])
            elif find(l, 'NEWDPF dihe0')>-1:
                angList = []
                for n in ll[3:]:
                    angList.append(float(n))
                d['torsion_values'] = angList
                d['num_torsions'] = len(angList)
            elif find(l, 'Intermol. vdW + Hbond Energy ')>-1:
                #AD4 specific model information:
                # USER         Intermol. vdW + Hbond Energy   =  -14.63 kcal/mol
                d['vdw_energy'] = float((ll)[7])
            elif find(l, 'Intermol. Electrostatic Energy')>-1:
                #USER         Intermol. Electrostatic Energy =   -0.62 kcal/mol
                d['estat_energy'] = float((ll)[5])
            elif find(l, '(3) Torsional Free Energy')>-1:
                #USER    (3) Torsional Free Energy           =   +3.84 kcal/mol
                d['torsional_energy'] = float((ll)[6])
            elif find(l, "(4) Unbound System's Energy")>-1:
                #USER    (4) Unbound System's Energy         =   -0.85 kcal/mol
                try:
                    d['unbound_energy'] = float((ll)[7])
                except:
                    d['unbound_energy'] = float((ll)[6])
            elif find(l, 'ATOM')>-1:
                coords.append([float(l[30:38]),float(l[38:46]),float(l[46:54])])
                try:
                    vdW.append(float(l[54:60]))
                except:
                    #print 'vdw:', l[54:60], ' RAISED!'
                    vdW.append(0.0)
                try:
                    Elec.append(float(l[60:66]))
                except:
                    #print 'estat:', l[60:66], ' RAISED!'
                    Elec.append(0.0)
                if len(l)>77:
                    try:
                        binding_energy2 = float(l[70:76])
                    except:
                        pass
            elif find(l, 'HETA')>-1:
                coords.append([float(l[30:38]),float(l[38:46]),float(l[46:54])])
                try:
                    vdW.append(float(l[54:60]))
                except:
                    vdW.append(0.0)
                try:
                    Elec.append(float(l[60:66]))
                except:
                    Elec.append(0.0)
            d['coords'] = coords
            d['vdw_energies'] = vdW
            d['estat_energies'] = Elec
            d['total_energies'] = Numeric.array(Numeric.array(vdW)+Numeric.array(Elec)).tolist()
            if binding_energy2 and not d.has_key('binding_energy'):
                d['binding_energy'] = binding_energy2
                d['docking_energy'] = binding_energy2
        return d






#!/usr/bin/env python
#
# 
#
# $Header: /opt/cvs/python/packages/share1.5/AutoDockTools/Utilities24/prepare_receptor4.py,v 1.14 2014/01/30 17:39:23 rhuey Exp $
#
import os 

from MolKit import Read
import MolKit.molecule
import MolKit.protein
from AutoDockTools.MoleculePreparation import AD4ReceptorPreparation


if __name__ == '__main__':
    import sys
    import getopt


    def usage():
        "Print helpful, accurate usage statement to stdout."
        print "Usage: prepare_receptor4.py -r filename"
        print
        print "    Description of command..."
        print "         -r   receptor_filename "
        print "        supported file types include pdb,mol2,pdbq,pdbqs,pdbqt, possibly pqr,cif"
        print "    Optional parameters:"
        print "        [-v]  verbose output (default is minimal output)"
        print "        [-o pdbqt_filename]  (default is 'molecule_name.pdbqt')"
        print "        [-A]  type(s) of repairs to make: "
        print "             'bonds_hydrogens': build bonds and add hydrogens "
        print "             'bonds': build a single bond from each atom with no bonds to its closest neighbor" 
        print "             'hydrogens': add hydrogens"
        print "             'checkhydrogens': add hydrogens only if there are none already"
        print "             'None': do not make any repairs "
        print "             (default is 'None')"
        print "        [-C]  preserve all input charges ie do not add new charges "
        print "             (default is addition of gasteiger charges)"
        print "        [-p]  preserve input charges on specific atom types, eg -p Zn -p Fe"
        print "        [-U]  cleanup type:"
        print "             'nphs': merge charges and remove non-polar hydrogens"
        print "             'lps': merge charges and remove lone pairs"
        print "             'waters': remove water residues"
        print "             'nonstdres': remove chains composed entirely of residues of"
        print "                      types other than the standard 20 amino acids"
        print "             'deleteAltB': remove XX@B atoms and rename XX@A atoms->XX"
        print "             (default is 'nphs_lps_waters_nonstdres') "
        print "        [-e]  delete every nonstd residue from any chain"
        print "              'True': any residue whose name is not in this list:"
        print "                      ['CYS','ILE','SER','VAL','GLN','LYS','ASN', "
        print "                      'PRO','THR','PHE','ALA','HIS','GLY','ASP', "
        print "                      'LEU', 'ARG', 'TRP', 'GLU', 'TYR','MET', "
        print "                      'HID', 'HSP', 'HIE', 'HIP', 'CYX', 'CSS']"
        print "              will be deleted from any chain. "
        print "              NB: there are no  nucleic acid residue names at all "
        print "              in the list and no metals. "
        print "             (default is False which means not to do this)"
        print "        [-M]  interactive "
        print "             (default is 'automatic': outputfile is written with no further user input)"
        print "        [-d dictionary_filename] file to contain receptor summary information"


    # process command arguments
    try:
        opt_list, args = getopt.getopt(sys.argv[1:], 'r:vo:A:Cp:U:eM:d:')

    except getopt.GetoptError, msg:
        print 'prepare_receptor4.py: %s' %msg
        usage()
        sys.exit(2)

    # initialize required parameters
    #-s: receptor
    receptor_filename =  None

    # optional parameters
    verbose = None
    #-A: repairs to make: add bonds and/or hydrogens or checkhydrogens
    repairs = ''
    #-C default: add gasteiger charges 
    charges_to_add = 'gasteiger'
    #-p preserve charges on specific atom types
    preserve_charge_types=None
    #-U: cleanup by merging nphs_lps, nphs, lps, waters, nonstdres
    cleanup  = "nphs_lps_waters_nonstdres"
    #-o outputfilename
    outputfilename = None
    #-m mode 
    mode = 'automatic'
    #-e delete every nonstd residue from each chain
    delete_single_nonstd_residues = None
    #-d dictionary
    dictionary = None

    #'r:vo:A:Cp:U:eMh'
    for o, a in opt_list:
        if o in ('-r', '--r'):
            receptor_filename = a
            if verbose: print 'set receptor_filename to ', a
        if o in ('-v', '--v'):
            verbose = True
            if verbose: print 'set verbose to ', True
        if o in ('-o', '--o'):
            outputfilename = a
            if verbose: print 'set outputfilename to ', a
        if o in ('-A', '--A'):
            repairs = a
            if verbose: print 'set repairs to ', a
        if o in ('-C', '--C'):
            charges_to_add = None
            if verbose: print 'do not add charges'
        if o in ('-p', '--p'):
            if not preserve_charge_types:
                preserve_charge_types = a
            else:
                preserve_charge_types = preserve_charge_types + ','+ a
            if verbose: print 'preserve initial charges on ', preserve_charge_types
        if o in ('-U', '--U'):
            cleanup  = a
            if verbose: print 'set cleanup to ', a
        if o in ('-e', '--e'):
            delete_single_nonstd_residues  = True
            if verbose: print 'set delete_single_nonstd_residues to True'
        if o in ('-M', '--M'):
            mode = a
            if verbose: print 'set mode to ', a
        if o in ('-d', '--d'):
            dictionary  = a
            if verbose: print 'set dictionary to ', dictionary
        if o in ('-h', '--'):
            usage()
            sys.exit()


    if not receptor_filename:
        print 'prepare_receptor4: receptor filename must be specified.'
        usage()
        sys.exit()

    #what about nucleic acids???

    mols = Read(receptor_filename)
    if verbose: print 'read ', receptor_filename
    mol = mols[0]
    preserved = {}
    has_autodock_element = False
    if charges_to_add is not None and preserve_charge_types is not None:
        if hasattr(mol, 'allAtoms') and not hasattr(mol.allAtoms[0], 'autodock_element'):
            file_name, file_ext = os.path.splitext(receptor_filename)
            if file_ext == '.pdbqt':
                has_autodock_element = True
        if preserve_charge_types is not None and has_autodock_element==False:
            print 'prepare_receptor4: input format does not have autodock_element SO unable to preserve charges on ' + preserve_charge_types
            print 'exiting...'
            sys.exit()  
        preserved_types = preserve_charge_types.split(',') 
        if verbose: print "preserved_types=", preserved_types
        for t in preserved_types:
            if verbose: print 'preserving charges on type->', t
            if not len(t): continue
            ats = mol.allAtoms.get(lambda x: x.autodock_element==t)
            if verbose: print "preserving charges on ", ats.name
            for a in ats:
                if a.chargeSet is not None:
                    preserved[a] = [a.chargeSet, a.charge]

    if len(mols)>1:
        if verbose: print "more than one molecule in file"
        #use the molecule with the most atoms
        ctr = 1
        for m in mols[1:]:
            ctr += 1
            if len(m.allAtoms)>len(mol.allAtoms):
                mol = m
                if verbose: print "mol set to ", ctr, "th molecule with", len(mol.allAtoms), "atoms"
    mol.buildBondsByDistance()

    if verbose:
        print "setting up RPO with mode=", mode,
        print "and outputfilename= ", outputfilename
        print "charges_to_add=", charges_to_add
        print "delete_single_nonstd_residues=", delete_single_nonstd_residues

    RPO = AD4ReceptorPreparation(mol, mode, repairs, charges_to_add, 
                        cleanup, outputfilename=outputfilename,
                        preserved=preserved, 
                        delete_single_nonstd_residues=delete_single_nonstd_residues,
                        dict=dictionary)    

    if charges_to_add is not None:
        #restore any previous charges
        for atom, chargeList in preserved.items():
            atom._charges[chargeList[0]] = chargeList[1]
            atom.chargeSet = chargeList[0]


# To execute this command type:
# prepare_receptor4.py -r pdb_file -o outputfilename -A checkhydrogens 

